#ifndef DATASTRUCTURES
#define DATASTRUCTURES


#include <QPointF>
#include <QLineF>
#include <armadillo>
#include <QVector>

/**
 * @brief Struktura, do které se ukládají nezpracované měření.
 */
struct measure_t {
    qreal angle; /**< Úhel v stupních. */
    qreal distance; /**< Vzdálenost v mm. */
    int quality; /**< Kvalita signálu (síla signálu při návratu). */
};

/**
 * @brief Struktura k ukládání jednoho souboru měření (celá otáčka).
 */
struct data_t {
    int measureCount; /**< Počet vzorků uložených v measure_t poli. */
    measure_t *measure; /**< Pole measure_t. */
};

/**
 * @brief Struktura k ukládání přímek.
 */
struct line_t {
    QLineF line; /**< Daná přímka, viz. QT dokumentace k QLineF. */
    qreal error; /**< Chyba přímky. */
};

/**
 * @brief Ukládání parametrů k přímce.
 */
struct lineParams_t {
    int indexStart; /**< Index prvního bodu přímky v poli bodů v area_t. */
    int indexEnd; /**< Index posledního bodu přímky v poli bodů v area_t. */
    qreal m; /**< Parametr přímky (k). */
    qreal b; /**< Parametr přímky (q). */
    qreal error; /**< Chyba přímky vydělená počtem bodů na přímce, takže se jedná o chybu na bod. */
    qreal fitsR2; /**< Koeficient určující přesnost aproximace v % (coefficient of determination). */
};

/**
 * @brief Struktura k ukládání průsečíku dvou přímek.
 */
struct intersections_t {
    int indexFirst; /**< Index první přímky line_t. */
    int indexSecond; /**< Index druhé přímky line_t. */
    qreal angle; /**< Úhel dvou přímek, které mají daný průsečík. */
    QPointF intersection; /**< Pozice průsečíku. */
};

/**
 * @brief Parametry průsečíku.
 */
struct intersectionParams_t {
    int indexFirst; /**< Index prvního průsečíku intersections_t. */
    int indexSecond; /**< Index druhého průsečíku intersections_t . */
    qreal len; /**< Vzdálenost těchto dvou průsečíků. */
};

/**
 * @brief Hlavní struktura aplikace. V této struktůře jsou uložena všechna naměřená data a představuje jednu otáčku skeneru.
 *
 * Všechny parametry z registrace závisí na proměnné transformable, která musí být TRUE.
 */
struct area_t {

    /**
     * @brief Kontruktor nastavující defaultní hodnoty.
     */
    area_t() {
        pointCount = 0;
        cornerCount = 0;
        lineCount = 0;
        sigma2 = 0;
        scale = 0;
        transformableToIndex = 0;
        nextAreaIndex = 0;
        intersectionCount = 0;
        transformable = false;
        landmarkFinished = false;
    }

    int pointCount; /**< Počet bodů v poli *points. */
    QPointF *points; /**< Pole bodů. */
    qreal *distanceCoef; /**< Pole distančních koeficientů, stejná velikost jako pole bodů. */
    int cornerCount; /**< Počet nalezených rohů v této area_t. */
    QPointF *corners; /**< Pole nalezených rohů. */
    int lineCount; /**< Počet nalezených přímek v této area_t. */
    QLineF *lines; /**< Pole přímek. */
    lineParams_t *lineParams; /**< Pole parametrů přímek. */
    int intersectionCount; /**< Počet uložených průsečíků. */
    intersections_t *intersections; /**< Pole průsečíků. */
    intersectionParams_t *inParams; /**< Pole parametrů průsečíků. */
    bool landmarkFinished; /**< Nastaveno na TRUE pokud je dokončené hledání důležitých objektů, jinak FALSE. */
    arma::mat rotationMatrix; /**< Rotační matice nalezené transformace, použitelná pouze pokud je transformable TRUE. */
    arma::vec translationVector; /**< Translační vektor nalezené transformace, taktéž použitelné pouze pokud je transformable TRUE. */
    double sigma2; /**< Hodnota kovariance nalezené transformace, při správném spojení musí být < 1e-4. */
    double scale; /**< Hodnota měřítka, při kterém registrační algoritmus skončil. */
    bool transformable; /**< Nastaveno na TRUE pokud je transformace možná, jinak FALSE. */
    int transformableToIndex; /**< Index area_t, ke které se tato oblast zarovnává. */
    int nextAreaIndex; /**< Index area_t, kterou lze zarovnat k této oblasti. */
};

/**
 * @brief Struktura k ukládání parametrů normalizace mračna bodů.
 */
struct normalization_t {
    double scale;
    arma::mat xd, yd;
};

/**
 * @brief Struktura k ukládáná pozice robota. Aktuální pozice je vždy [0, 0].
 */
struct robotPosition_t {
    /**
     * @brief Kontruktor, inicializace výchozí pozice.
     */
    robotPosition_t() {
        position = arma::zeros<arma::mat>(1, 2);
        zero = arma::zeros<arma::mat>(1, 2);
    }

    /**
     * @brief Přidá aktuální pozici do historie a její hodnotu aktualizuje
     * @param area - Oblast, která určuje aktuální pozici spolu s rotační maticí a translačním vektorem.
     * @param start - Index bodu, kde v poli bodů mapy začíná tato oblast.
     */
    void addNext(area_t *area, int start) {
        firstPoint.append(start);
        position.resize(position.n_rows + 1, 2);
        position = position * area->rotationMatrix - repmat(area->translationVector.t(), position.n_rows, 1);
        position.submat(position.n_rows - 1, 0, position.n_rows - 1, 1) = zero;
    }

    arma::mat position;
    QVector<int> firstPoint;

private:
    arma::mat zero;
};

/**
 * @brief Struktura k ukládání důležitých objektů jako rohy a hrany.
 */
struct landmark_t {
    landmark_t() {
        this->truthFactor = 0;
        this->type = 0;
    }

    qreal truthFactor; /**< Důvěryhodnost objektu. */
    int type; /**< Typ objektu: 0 - roh (corner), 1 - přímka (line). */
};


#endif // DATASTRUCTURES

