#include "landmarkextraction.h"

#include <qmath.h>
#include <QDebug>

LandmarkExtraction::LandmarkExtraction()
{
    this->arrayIndex = 100;
    this->cornerCount = 0;
    this->corners = new QPointF[this->arrayIndex];
    this->maxAngle = 2.5;
    this->minAngle = 0.697;

    this->lineCount = 0;
    this->lineArrayCount = 100;
    this->linePointStep = 2;
}


/**
 * @brief Extrahování rohů pomocí sliding window algoritmu.
 * @param area - V této oblasti se hledá důležité body.
 *
 * Úhly mezi, kterými se hledá roh (minAngle, maxAngle). Distanční koeficient se používá k odhalení jestli body leží daleko od sebe.
 */
void LandmarkExtraction::cornerExtraction(area_t &area)
{
    qreal angle;
    qreal vlen = 0;
    qreal ulen = 0;
    int cornerCount = 0;
    int precisionPoint = 5;
    bool isCorner = false;

    for(int i = precisionPoint; i < area.pointCount - precisionPoint; i++) {
        this->angleTwoVectors(area.points[i-precisionPoint], area.points[i+precisionPoint], area.points[i], angle);
        if(angle < maxAngle && angle > minAngle) {
            isCorner = true;
            // really corner?
            for(int h = precisionPoint-1; h > 0; h--) {
                qreal maxAngleModified = maxAngle;
                qreal minAngleModified = minAngle;
                this->angleTwoVectors(area.points[i-h], area.points[i+h], area.points[i], angle, vlen, ulen);
                qreal uLenForPoint = ulen / (h+1);
                qreal vLenForPoint = vlen / (h+1);
                if(angle > maxAngleModified || angle < minAngleModified || uLenForPoint > area.distanceCoef[i]*1.5 ||
                    vLenForPoint > area.distanceCoef[i]*1.5) {
                    isCorner = false;
                }
            }
            if(isCorner) {
                corners[cornerCount] = area.points[i];
                cornerCount++;
                if(cornerCount >= this->arrayIndex) {
                    this->resizeCornerArray();
                }
            }
        }
    }

    area.cornerCount = cornerCount;
    area.corners = new QPointF[cornerCount];
    for(int i = 0; i < cornerCount; i++) {
        area.corners[i] = corners[i];
    }
}

/**
 * @brief Realokace pole k ukládání rohů.
 * @param addSize - Velikost o kterou se pole realokuje, defaultní hodnota je 50.
 */
void LandmarkExtraction::resizeCornerArray(int addSize)
{
    QPointF *oldCorners = corners;
    this->arrayIndex += addSize;
    this->corners = new QPointF[this->arrayIndex];
    for(int i = 0; i < this->arrayIndex - addSize; i++) {
        corners[i] = oldCorners[i];
    }
}

/**
 * @brief Úhel mezi dvěma vektory danými třemi body.
 * @param first - První bod (začátek prvního vektoru).
 * @param second - Druhý bod (začátek druhého vektoru).
 * @param middle - Koncový bod obou vektorů.
 * @param angle - Vypočítaný úhel.
 */
void LandmarkExtraction::angleTwoVectors(QPointF &first, QPointF &second, QPointF &middle, qreal &angle)
{
    QPointF u = QPointF(first.x() - middle.x(), first.y() - middle.y());
    QPointF v = QPointF(second.x() - middle.x(), second.y() - middle.y());
    qreal ulen = qSqrt(qPow(u.x(), 2) + qPow(u.y(), 2));
    qreal vlen = qSqrt(qPow(v.x(), 2) + qPow(v.y(), 2));

    angle = qAcos((u.x() * v.x() + u.y() * v.y())/(ulen * vlen));
}


/**
 * @brief Úhel mezi dvěma vektory danými třemi body.
 * @param first - První bod (začátek prvního vektoru).
 * @param second - Druhý bod (začátek druhého vektoru).
 * @param middle - Koncový bod obou vektorů.
 * @param angle - Vypočítaný úhel.
 * @param vl - Vypočítaná délka prvního vektoru.
 * @param ul - Vypočítaná délka druhého vektoru.
 */
void LandmarkExtraction::angleTwoVectors(QPointF &first, QPointF &second, QPointF &middle, qreal &angle, qreal &vl, qreal &ul)
{
    QPointF u = QPointF(first.x() - middle.x(), first.y() - middle.y());
    QPointF v = QPointF(second.x() - middle.x(), second.y() - middle.y());
    qreal ulen = qSqrt(qPow(u.x(), 2) + qPow(u.y(), 2));
    qreal vlen = qSqrt(qPow(v.x(), 2) + qPow(v.y(), 2));

    angle = qAcos((u.x() * v.x() + u.y() * v.y())/(ulen * vlen));
    vl = vlen;
    ul = ulen;
}

/**
 * @brief Délka vektoru daného dvěma body.
 * @param first - První bod vektoru.
 * @param second - Druhý bod vektoru.
 * @param len - Vypočítaná délka.
 */
void LandmarkExtraction::vectorLength(QPointF &first, QPointF &second, qreal &len)
{
    QPointF u = QPointF(first.x() - second.x(), first.y() - second.y());
    len = qSqrt(qPow(u.x(), 2) + qPow(u.y(), 2));
}

/**
 * @brief Funkce k přidávání nových hran (přímek) k oblasti area_t (y = mx + b).
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index druhého bodu přímky.
 * @param m - Parametr přímky.
 * @param b - Parametr přímky.
 * @param error - Chyba přímky vydělená počtem bodů (počtem bodů na této přímce). Takže se jedná o vhybu na bod.
 * @param fitsR2 - Určuje jak dobrá je aproximace přímky v procentech.
 * @param area - Oblast do které bude přímka přidána.
 */
void LandmarkExtraction::addNewLine(int indexStart, int indexEnd, qreal m, qreal b, qreal error, qreal fitsR2, area_t &area)
{
    area.lines[area.lineCount].setPoints(area.points[indexStart], area.points[indexEnd]);
    area.lineParams[area.lineCount].indexStart = indexStart;
    area.lineParams[area.lineCount].indexEnd = indexEnd;
    area.lineParams[area.lineCount].m = m;
    area.lineParams[area.lineCount].b = b;
    area.lineParams[area.lineCount].error = error;
    area.lineParams[area.lineCount].fitsR2 = fitsR2;
    area.lineCount++;
}

/**
 * @brief Funkce k editaci hran (přímek).
 *
 * Pro více informací viz. addNewLine().
 *
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index druhého bodu přímky.
 * @param m - Parametr přímky.
 * @param b - Parametr přímky.
 * @param error - Chyba přímky vydělená počtem bodů (počtem bodů na této přímce). Takže se jedná o vhybu na bod.
 * @param fitsR2 - Určuje jak dobrá je aproximace přímky v procentech.
 * @param area - Oblast do které bude přímka přidána.
 * @param lineIndex - Index přímky, která má být editována.
 */
void LandmarkExtraction::editExistingLine(int indexStart, int indexEnd, qreal m, qreal b, qreal error, qreal fitsR2, area_t &area, int lineIndex)
{
    area.lines[lineIndex].setPoints(area.points[indexStart], area.points[indexEnd]);
    area.lineParams[lineIndex].indexStart = indexStart;
    area.lineParams[lineIndex].indexEnd = indexEnd;
    area.lineParams[lineIndex].m = m;
    area.lineParams[lineIndex].b = b;
    area.lineParams[lineIndex].error = error;
    area.lineParams[lineIndex].fitsR2 = fitsR2;
}

/**
 * @brief Funkce začínající hledání přímek v area_t.
 * @param area - Oblast, v které budou přímky hledány.
 */
void LandmarkExtraction::lineFitting(area_t &area)
{
    area.lineCount = 0;
    area.lines = new QLineF[100];
    area.lineParams = new lineParams_t[100];
    int len = area.pointCount > linePointStep ? linePointStep : area.pointCount;
    qreal error = 0;
    this->lineExtract(0, len, area, error);
}

/**
 * @brief Hlavní funkce k hledání přímek, jsou zde hledány i průsečíky a jejich parametry.
 * Průsečíky jsou zatim pouze pro testování (ještě se nevyužívají). Pokud je nalezená aproximace přímky dostatečně přesna, přidá se do pole, ale pokud je chyba velká, volá se funkce lineEliminateError().
 * @param indexStart - Index bodu, od kterého se má začít (pro všechny naměřené body = 0).
 * @param indexEnd - Index koncového bodu přímky (pro začátek je to indexStart + 2).
 * @param area - Oblast s naměřenými body, kde jsou uloženy i přímky.
 * @param error - Výchozí chyba.
 */
void LandmarkExtraction::lineExtract(int indexStart, int indexEnd, area_t &area, qreal &error)
{
    qreal m = 0;
    qreal b = 0;
    qreal sErrorY = 0;
    qreal r2 = 0;

    indexStart = 0;
    indexEnd = -1;

    // searching for lines
    while(indexEnd < area.pointCount) {
        indexEnd++;
        this->getNextPossibleLine(indexStart, indexEnd, area);
        if(indexEnd - indexStart > 6) {
            this->getLineParams(indexStart, indexEnd, area, m, b, sErrorY);
            this->calculateSquareError(area, indexStart, indexEnd, m, b, sErrorY, error, r2);
            error = error < 0 ? error * -1 : error;
            error = qSqrt(error)/(indexEnd-indexStart);
            if(r2 < 0.98 || error > 3) {
                this->lineEliminateError(indexStart, indexEnd, 0.98, area);
            } else {
                this->addNewLine(indexStart, indexEnd, m, b, error, r2, area);
            }
        }
        indexStart = indexEnd;
    }

    qreal x;
    qreal y;
    qreal m1m2;


    // intersections

    int interCountAll = 0.5 * area.lineCount * (area.lineCount - 1);
    area.intersectionCount = interCountAll;
    area.intersections = new intersections_t[interCountAll];
    int intersectionsCount = 0;
    qreal angle = 0;

    for(int i = 0; i < area.lineCount; i++) {
        //qDebug() << i << "errorOnPoint" << area.lineParams[i].error << "fits" << area.lineParams[i].fitsR2;
        for(int h = i+1; h < area.lineCount; h++) {
            m1m2 = area.lineParams[i].m - area.lineParams[h].m;
            if(m1m2 != 0) {
                x = (area.lineParams[h].b - area.lineParams[i].b)/(m1m2);
                y = area.lineParams[i].m * x + area.lineParams[i].b;
                //qDebug() << "line " << i << "(b = " << area.lineParams[i].b << ", m = " << area.lineParams[i].m << ")" << "_"
                //         << h << "(b = " << area.lineParams[h].b << ", m = " << area.lineParams[h].m << ")" << "_"
                //         << "point [" << x << "," << y << "]";
                area.intersections[intersectionsCount].indexFirst = i;
                area.intersections[intersectionsCount].indexSecond = h;
                area.intersections[intersectionsCount].intersection.setX(x);
                area.intersections[intersectionsCount].intersection.setY(y);
                QPointF first = area.lines[i].p1();
                QPointF second = area.lines[h].p1();
                this->angleTwoVectors(first, second, area.intersections[intersectionsCount].intersection, angle);
                area.intersections[intersectionsCount].angle = angle;
                intersectionsCount++;
                //qDebug() << "angle= " << angle;
            }
        }
    }


    // intersection params

    int interParamCountAll = 0.5 * interCountAll * (interCountAll - 1);
    area.inParams = new intersectionParams_t[interParamCountAll];
    int interParamCount = 0;
    qreal len = 0;

    for(int i = 0; i < interCountAll; i++) {
        for(int h = i+1; h < interCountAll; h++) {
            this->vectorLength(area.intersections[i].intersection, area.intersections[h].intersection, len);
            area.inParams[interParamCount].len = len;
            area.inParams[interParamCount].indexFirst = i;
            area.inParams[interParamCount].indexSecond = h;
            interParamCount++;
        }
    }

    area.landmarkFinished = true;
}

/**
 * @brief Hledá další možnou přímku, začínající od předchozí tedy na indexEnd. Indexy jsou editovány dle další nalezené přímky.
 * @param indexStart - Index prvního bodu předchozí přímky (je přepsán).
 * @param indexEnd - Index posledního bodu předchozí přímky (je přepsán).
 * @param area - Oblast s naměřenými body, kde jsou uloženy i přímky.
 */
void LandmarkExtraction::getNextPossibleLine(int &indexStart, int &indexEnd, area_t &area)
{
    qreal len = 0;
    int count = 0;
    qreal distanceCoef = 0;
    for(int i = indexEnd+1; i < area.pointCount; i++) {
        this->vectorLength(area.points[i-1], area.points[i], len);
        distanceCoef = (area.distanceCoef[i-1] + area.distanceCoef[i])/2;
        if(len > distanceCoef*2) {
            indexStart = indexEnd;
            indexEnd = i-1;
            return;
        }
        count++;
    }
    indexStart = indexEnd;
    indexEnd += count;
}

/**
 * @brief Funkce snažící se snižit chybu přímky tím, že ji rozdělí na více přímek, stejně jako split and merge algoritmus.
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param fits - Jak dobrá musí být výsledná aproximace r2 (teď nepoužito).
 * @param area - Oblast s naměřenými body, kde jsou uloženy i přímky.
 */
void LandmarkExtraction::lineEliminateError(int indexStart, int indexEnd, qreal fits, area_t &area)
{
    int start = indexStart;
    int end = indexStart+2;
    int actualEnd = 0;
    qreal m = 0;
    qreal b = 0;
    qreal error = 0;
    qreal prevError = 0;
    qreal sErrorY = 0;
    qreal r2 = 0;
    qreal angle = 0;

    for(int i = indexStart; i <= indexEnd; i++) {
        this->getLineParams(start, end, area, m, b, sErrorY);
        this->calculateSquareError(area, start, end, m, b, sErrorY, error, r2);
        this->angleTwoVectors(area.points[start], area.points[end], area.points[end-1], angle);
        error = error < 0 ? error * -1 : error;
        error = qSqrt(error)/(end-start);
        if(error > 3 || angle > 3.2 || angle < 2.355) {
            if(end - start > 6) {
                this->addNewLine(start, end-1, m, b, prevError, r2, area);
                this->searchForMoreAcurateLine(start, end-1, area, r2, prevError, actualEnd);
            }
            start = end;
        }
        end++;
        actualEnd = end;
        prevError = error;
    }
}

/**
 * @brief Funkce zpřesňuje nalezenou přímku tím, že zkouší odebrat bod ze začátku nebo z konce.
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param area - Oblast s naměřenými body, kde jsou uloženy i přímky.
 * @param r2 - Aktuální přesnost aproximace v procentech (pokud je přímka upravena, hodnota je změněna).
 * @param error - Aktuální chyba přímky (pokud je přímka upravena, hodnota je změněna).
 * @param actualEnd - Výsledný index koncového bodu přímky. Pokud se přímka nezlepší a zůstává stejná, hodnota se rovná indexEnd.
 */
void LandmarkExtraction::searchForMoreAcurateLine(int indexStart, int indexEnd, area_t &area, qreal &r2, qreal &error, int &actualEnd)
{
    qreal m = 0;
    qreal b = 0;
    qreal sErrorY = 0;
    qreal actualR2 = r2;
    qreal actualError = error;

    int nextStart = indexStart + 1;
    int preEnd = indexEnd - 1;
    actualEnd = indexEnd;

    // line -1 point from start
    this->getLineParams(nextStart, indexEnd, area, m, b, sErrorY);
    this->calculateSquareError(area, nextStart, indexEnd, m, b, sErrorY, error, r2);
    error = error < 0 ? error * -1 : error;
    error = qSqrt(error)/(indexEnd-indexStart);
    if(error < actualError) {
        actualR2 = r2;
        actualError = error;
        this->editExistingLine(nextStart, indexEnd, m, b, error, r2, area, area.lineCount-1);
    }

    // line -1 point from end
    this->getLineParams(indexStart, preEnd, area, m, b, sErrorY);
    this->calculateSquareError(area, indexStart, preEnd, m, b, sErrorY, error, r2);
    error = error < 0 ? error * -1 : error;
    error = qSqrt(error)/(indexEnd-indexStart);
    if(error < actualError) {
        //actualR2 = r2;
        //actualError = error;
        this->editExistingLine(indexStart, preEnd, m, b, error, r2, area, area.lineCount-1);
        actualEnd = preEnd;
    }
}

/**
 * @brief Vypočítá parametry přímky.
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param area - Oblast, v které jsou uložené body a přímky.
 * @param theta - Úhel přímky.
 * @param ro - Vzdálenost přímky od počátku [0, 0].
 */
void LandmarkExtraction::getLineParams(int indexStart, int indexEnd, area_t &area, qreal &theta, qreal &ro)
{
    qreal xm = 0;
    qreal ym = 0;
    int count = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        xm += area.points[i].x();
        ym += area.points[i].y();
        count++;
    }

    xm = xm / count;
    ym = ym / count;

    qreal top = 0;
    qreal bottom = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        top += (ym - area.points[i].y()) * (xm - area.points[i].x());
        bottom += qPow(ym - area.points[i].y(), 2) - qPow(xm - area.points[i].x(), 2);
    }

    theta = 0.5 * qAtan((-2*top)/bottom);
    ro = xm * qCos(theta) + ym * qSin(theta);
}

/**
 * @brief Vypočítá parametry přímky (y = mx + b).
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param area - Oblast, v které jsou uložené body a přímky.
 * @param m - Parametr přímky.
 * @param b - Parametr přímky.
 * @param sErrorY - Vypočtená kvadratická chyba přímky.
 */
void LandmarkExtraction::getLineParams(int indexStart, int indexEnd, area_t &area, qreal &m, qreal &b, qreal &sErrorY)
{
    qreal xm = 0;
    qreal ym = 0;
    qreal xy = 0;
    qreal x2 = 0;
    int count = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        xm += area.points[i].x();
        ym += area.points[i].y();
        xy += area.points[i].x() * area.points[i].y();
        x2 += qPow(area.points[i].x(), 2);
        count++;
    }

    xm = xm / count;
    ym = ym / count;
    xy = xy / count;
    x2 = x2 / count;

    m = (xm * ym - xy)/(qPow(xm, 2) - x2);
    b = ym - m * xm;

    sErrorY = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        sErrorY += qPow(area.points[i].y() - ym, 2);
    }
}

/**
 * @brief Kvadratická chyba vypočítaná z přímky, která je dána vzdáleností ro a úhlem theta.
 * @param area - Oblast, v které jsou uložené body a přímky.
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param ro - Parametr přímky.
 * @param theta - Parametr přímky.
 * @param error - Vypočtená kvadratická chyba přímky.
 */
void LandmarkExtraction::calculateSquareError(area_t &area, int indexStart, int indexEnd, qreal &ro, qreal &theta, qreal &error)
{
    error = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        error += qPow(ro - (area.points[i].x() * qCos(theta) + area.points[i].y() * qSin(theta)), 2);
    }
}

/**
 * @brief Kvadratická chyba vypočítaná z přímky, která je dána rovnicí y = mx + b.
 * @param area - Oblast, v které jsou uložené body a přímky.
 * @param indexStart - Index prvního bodu přímky.
 * @param indexEnd - Index posledního bodu přímky.
 * @param m - Parametr přímky.
 * @param b - Parametr přímky.
 * @param sErrorY - Chyba přímky vypočtena ve funkci getLineParams().
 * @param error - Výsledná kvadratická chyba.
 * @param r2 - Vypočtená míra v procentech jak dobrá je aproximace (coefficient of determination).
 */
void LandmarkExtraction::calculateSquareError(area_t &area, int indexStart, int indexEnd, qreal &m, qreal &b, qreal &sErrorY, qreal &error, qreal &r2)
{
    error = 0;
    for(int i = indexStart; i <= indexEnd; i++) {
        error += qPow((area.points[i].y() - (m * area.points[i].x() + b)), 2);
    }
    r2 = 1 - error / sErrorY;
}

/**
 * @brief Hledání společných objektů ve dvou oblastech. Ještě není dokončeno, pouze pro testování.
 * @param areaFirst - První oblast.
 * @param areaSecond - Druhá oblast.
 */
void LandmarkExtraction::searchForConnections(area_t &areaFirst, area_t &areaSecond)
{
    qreal sizeHalf = 100;
    qreal eval = 0, e, x1, y1, x2, y2, len;
    QVector<int> indexI;
    QVector<int> indexH;
    QVector<QPair<QPointF, QPointF>> connectedPoints;
    for(int i = 0; i < areaFirst.cornerCount; i++) {
        for(int h = 0; h < areaSecond.cornerCount; h++) {
            if(!indexI.contains(i) && !indexH.contains(h)) {
                x1 = areaFirst.corners[i].x();
                y1 = areaFirst.corners[i].y();
                x2 = areaSecond.corners[h].x();
                y2 = areaSecond.corners[h].y();
                this->vectorLength(areaFirst.corners[i], areaSecond.corners[h], len);

                if(qRound(x1) == qRound(x2) && qRound(y1) == qRound(y2)) {
                    indexI.append(i);
                    indexH.append(h);
                    connectedPoints.append(QPair<QPointF, QPointF>(areaFirst.corners[i], areaSecond.corners[h]));
                    qDebug() << "len: " << len;
                    //eval++;
                }
                e = this->getPointsEvaluation(x1, y1, x2, y2, sizeHalf);
                if(e > 0 && len < sizeHalf) {
                    indexI.append(i);
                    indexH.append(h);
                    connectedPoints.append(QPair<QPointF, QPointF>(areaFirst.corners[i], areaSecond.corners[h]));
                    qDebug() << "len: " << len;
                }
                //qDebug() << "e: " << e;
                eval += e;
            }
        }
    }
    indexI.clear();
    indexH.clear();
    for(int i = 0; i < areaFirst.intersectionCount; i++) {
        for(int h = 0; h < areaSecond.intersectionCount; h++) {
            if(!indexI.contains(i) && !indexH.contains(h)) {
                if(areaFirst.intersections[i].angle > 0.1745 && areaFirst.intersections[i].angle < 2.967 &&
                        areaSecond.intersections[h].angle > 0.1745 && areaSecond.intersections[h].angle < 2.967) {
                    x1 = areaFirst.intersections[i].intersection.x();
                    y1 = areaFirst.intersections[i].intersection.y();
                    x2 = areaSecond.intersections[h].intersection.x();
                    y2 = areaSecond.intersections[h].intersection.y();
                    this->vectorLength(areaFirst.intersections[i].intersection, areaSecond.intersections[h].intersection, len);

                    if(qRound(x1) == qRound(x2) && qRound(y1) == qRound(y2)) {
                        indexI.append(i);
                        indexH.append(h);
                        connectedPoints.append(QPair<QPointF, QPointF>(areaFirst.intersections[i].intersection, areaSecond.intersections[i].intersection));
                        qDebug() << "len: " << len;
                        //eval++;
                    }
                    e = this->getPointsEvaluation(x1, y1, x2, y2, sizeHalf);
                    if(e > 0 && len < sizeHalf) {
                        indexI.append(i);
                        indexH.append(h);
                        connectedPoints.append(QPair<QPointF, QPointF>(areaFirst.intersections[i].intersection, areaSecond.intersections[i].intersection));
                        qDebug() << "len: " << len;
                    }
                    //qDebug() << "e: " << e;
                    eval += e;
                }
            }
        }
    }
    for(int i = 0; i < connectedPoints.length(); i++) {
        for(int h = 0; h < connectedPoints.length(); h++) {
            // angle between these lines
        }
    }
    qDebug() << "evaluation = " << eval << connectedPoints.length();
}

/**
 * @brief Ohodnocení bodů z různých oblastí. Ještě nedokončeno, pouze pro testování.
 * @param x1 - Souřadnice x prvního bodu.
 * @param y1 - Souřadnice y prvního bodu.
 * @param x2 - Souřadnice x druhého bodu.
 * @param y2 - Souřadnice y druhého bodu.
 * @param sizeHalf - Velikost okolí kolem bodu, zatím nepoužito.
 * @return
 */
qreal LandmarkExtraction::getPointsEvaluation(qreal &x1, qreal &y1, qreal &x2, qreal &y2, qreal &sizeHalf)
{
    qreal eval = 0;
    qreal size = sizeHalf;
    for(int i = 0; i < 5; i++) {
        if(x1 + size > x2 && x1 - size < x2 && y1 + size > y2 && y1 - size < y2) {
            eval += qPow(2, i);
        }
        size -= size / 2;
    }
    if(qRound(x1) == qRound(x2) && qRound(y1) == qRound(y2)) {
        eval += 1000;
    }
    return eval;
}
