#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "submappostproc.h"

#include <QFileDialog>
#include <QStandardItemModel>
#include <QtConcurrent/QtConcurrent>
#include <QFutureWatcher>
#include <QDebug>

/**
 * @brief Funkce k načítání dat ze souboru, spouští se ve vlastním vlákně.
 * @param dataReader - Třída slouži k načtení dat.
 * @param path - Cesta k souboru.
 * @param dataPreProcessor - Do této třídy se ukládá všechny načtené měření a provádí preprocesing.
 */
void fileToDataReader(DataReader *dataReader, QString &path, DataPreProcessor *dataPreProcessor)
{
    dataReader = new DataReader(path);
    for(int i = 0; i < dataReader->dataCount; i++) {
        dataPreProcessor->addNewArea(dataReader->data[i]);
    }
}

/**
 * @brief Funkce k načítání submap ze souboru.
 * @param postProc - Třída, která načítá a ukládá submapy.
 * @param path - Cesta k souboru.
 */
void loadSubmap(SubmapPostProc *postProc, QString &path)
{
    postProc->loadSubmap(path);
}

/**
 * @brief Funkce k zarovnání dvou mračen bodů.
 * @param preProcessor - Obsahuje mračna bodů.
 * @param statistics - Třída zobrazující statistiky.
 * @param staticIndex - Index prvního mračna bodů.
 * @param areaIndex - Index druhého mračna bodů.
 * @param areaTrans
 */
void twoPointSetsRegistrationOneStep(DataPreProcessor *preProcessor, Statistics *statistics, int staticIndex, int areaIndex, area_t *areaTrans)
{
    area_t *areaChanged = new area_t();
    qDebug() << "startAlg transforming=" << areaIndex << ", to=" << staticIndex;
    PointSetRegistration *reg = new PointSetRegistration();
    reg->setAreas(preProcessor->areas[staticIndex], preProcessor->areas[areaIndex]);
    reg->start(*areaChanged);
    if(preProcessor->areas[areaIndex]->sigma2 <= 5e-4f && preProcessor->areas[areaIndex]->scale > 0.95) {
        preProcessor->areas[areaIndex]->transformable = true;
        preProcessor->areas[areaIndex]->transformableToIndex = staticIndex;
        preProcessor->areas[staticIndex]->nextAreaIndex = areaIndex;
        statistics->setCouldTransform(true);
    } else {
        preProcessor->areas[areaIndex]->transformable = false;
        statistics->setCouldTransform(false);
    }
    statistics->setSigma(preProcessor->areas[areaIndex]->sigma2);

    if(!preProcessor->areas[areaIndex]->landmarkFinished) {
        LandmarkExtraction *landmarkExtraction = new LandmarkExtraction();
        landmarkExtraction->cornerExtraction(*preProcessor->areas[areaIndex]);
        preProcessor->areas[areaIndex]->landmarkFinished = true;
    }
    //qDebug() << "endAlg";
}

/**
 * @brief Extrahování důležitých objektů (samostatně, při zarovnávání se provádí automaticky).
 * @param preProcessor - Obsahuje mračna bodů.
 * @param index - Index mračna bodů.
 * @param corners - Jestli se mají extrahovat rohy.
 * @param lines - Jestli se mají extrahovat hrany.
 * @param indexPrev - Předchozí index mračna bodů, používano pouze pro testování funkce searchForConnections().
 */
void landmarkExtractionThread(DataPreProcessor *preProcessor, int index, bool corners, bool lines, int indexPrev)
{
    LandmarkExtraction *landmarkExtraction = new LandmarkExtraction();
    if(corners) {
        landmarkExtraction->cornerExtraction(*preProcessor->areas[index]);
    }
    if(lines) {
        landmarkExtraction->lineFitting(*preProcessor->areas[index]);
    }
    if(index != 0 && indexPrev != -1) {
        //landmarkExtraction->searchForConnections(preProcessor->areas[index], preProcessor->areas[indexPrev]);
    }
}

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::StaticRender, false);
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::RealTimeRender, false);
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::ProcessSubmap, true);
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::Settings, true);

    ui->toolButtonRegistration->setCheckable(true);
    ui->toolButtonShowLandmark->setCheckable(true);
    ui->toolButtonActualScan->setCheckable(true);
    ui->toolButtonMap->setCheckable(true);
    ui->toolButtonRealActual->setCheckable(true);
    ui->toolButtonRealMap->setCheckable(true);

    ui->toolButtonNext->setEnabled(false);
    ui->toolButtonStart->setEnabled(false);

    ui->toolButtonCheckSubmap->setHidden(true);

    ui->toolButtonActualScan->setChecked(true);
    ui->toolButtonRealActual->setChecked(true);
    ui->stackedWidget->setCurrentIndex(StackedWidget::ActualScan);
    ui->tabWidgetRightCol->setCurrentIndex(RightColumnTabs::StatisticsTab);
    ui->mainMenuTabs->setCurrentIndex(MainMenuTabs::File);

    connect(ui->toolButtonRealActual, SIGNAL(clicked(bool)), this, SLOT(on_toolButtonActualScan_clicked()));
    connect(ui->toolButtonRealMap, SIGNAL(clicked(bool)), this, SLOT(on_toolButtonMap_clicked()));

    this->dataPreProcessor = new DataPreProcessor();
    this->activeStaticRender = 0;

    this->landmarkExtraction = new LandmarkExtraction();

    connect(ui->toolButtonZoomIn, SIGNAL(clicked(bool)), this, SLOT(zoomIn()));
    connect(ui->toolButtonZoomOut, SIGNAL(clicked(bool)), this, SLOT(zoomOut()));

    connect(ui->renderAreaScan, SIGNAL(statistisUpdate()), this, SLOT(renderStatistics()));
    connect(ui->renderAreaMap, SIGNAL(statistisUpdate()), this, SLOT(renderStatistics()));

    activated = false;
    applyRegistration = false;
    showLandmark = false;
    completedFirstThreat = false;
    completedSecondThreat = false;
    renderSubmap = true;
    correctionRunning = false;
    this->obDraw = false;

    timer = new QTimer();
    connect(timer, SIGNAL(timeout()), this, SLOT(nextScan()));
    connect(timer, SIGNAL(timeout()), this, SLOT(checkRegistrationResults()));
    timer->start(500);

    areaChanged = new area_t();
    areaChanged->pointCount = 0;

    this->areaIndex = 0;
    this->connectedIndex = 1;
    this->connectedWith = 2;
    this->correctionCount = 0;
    this->forbiddenIndexes = QVector<int>(100);

    this->robotPos = new robotPosition_t();

    this->statistics = new Statistics();
    connect(statistics, SIGNAL(statisticsChanged()), this, SLOT(reloadStatistics()));
    connect(ui->tabWidgetRightCol, SIGNAL(currentChanged(int)), this, SLOT(reloadStatistics()));

    this->statisticsMap = new Statistics();
    connect(statisticsMap, SIGNAL(statisticsChanged()), this, SLOT(reloadStatistics()));

    this->statisticsMapRender = new Statistics();

    this->map = new MapClass(this->dataPreProcessor->areas, this->statistics, true);
    this->mapRender = new MapClass(this->dataPreProcessor->areas, this->statisticsMapRender, false);
    this->mapRender->setMaxDistance(1500);
    this->renderedSubmap = 0;
    connect(this->map, SIGNAL(updated()), this, SLOT(updateMapRender()));
    connect(this->map, SIGNAL(submapChanged()), this, SLOT(reloadSubmapTab()));

    this->mapProcess = new MapProcess();

    this->postProc = new SubmapPostProc();
    connect(this->postProc, SIGNAL(submapLoaded()), this, SLOT(reloadLoadedTable()));

    this->client = new UdpClient(this->dataPreProcessor);
    connect(client, SIGNAL(stateChange()), this, SLOT(clientChangeState()));
    connect(client, SIGNAL(statisticChange()), this, SLOT(reloadLidarTable()));
    connect(ui->toolButtonRealTimeStart, SIGNAL(clicked(bool)), client, SLOT(startPreprocessingData()));
    connect(ui->toolButtonRealTimeStop, SIGNAL(clicked(bool)), client, SLOT(stopPreprocessingData()));
    connect(ui->pushButtonGrab, SIGNAL(clicked(bool)), client, SLOT(startPreprocessingData()));
    connect(ui->pushButtonStop, SIGNAL(clicked(bool)), client, SLOT(stopPreprocessingData()));

    ui->stackedWidgetLidar->setCurrentIndex(StackedWidgetLidar::NotConnected);

    connect(ui->tabWidgetRightCol, SIGNAL(currentChanged(int)), this, SLOT(reloadSubmapTab()));
    connect(ui->tabWidgetRightCol, SIGNAL(currentChanged(int)), this, SLOT(reloadLoadedTable()));
    connect(ui->tableViewSubmaps, SIGNAL(clicked(QModelIndex)), this, SLOT(onSubmapClicked(QModelIndex)));
    connect(ui->mainMenuTabs, SIGNAL(currentChanged(int)), this, SLOT(checkMainMenuEnabled()));
}

MainWindow::~MainWindow()
{
    delete ui;
}

/**
 * @brief Nastavení tabulkového zobrazení.
 * @param tableView - Pro kterou tabulku se má nastavení provést.
 */
void MainWindow::setupTableView(QTableView *tableView) {
    tableView->setColumnWidth(0, 175);
    tableView->setColumnWidth(1, 175);
    tableView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
}

/**
 * @brief Zapnutí progress baru.
 */
void MainWindow::enableProgressBar()
{
    ui->progressBar->setMinimum(0);
    ui->progressBar->setMaximum(0);
}

/**
 * @brief Vypnutí progress baru.
 */
void MainWindow::disableProgressBar()
{
    ui->progressBar->setMaximum(100);
}

/**
 * @brief Po kliku na tlačítko openFile se otevře dialog s výběrem souboru.
 *
 * Po výběru se spustí funkce fileToDataReader() ve vlastním vlákně a načte statická data. Také se vlákno napojí pomocí connect() na signál finished() a na SLOT dataLoaded(), to znamená, že po dokončení se zavolá funkce dataLoaded().
 */
void MainWindow::on_toolButtonOpenFile_clicked()
{
    QString path = QFileDialog::getOpenFileName();
    if(!path.isEmpty()) {
        QStringList splitted = path.split("/");
        ui->toolButtonOpenFile->setEnabled(false);
        //run thread for data loading + pre-processing

        this->enableProgressBar();

        this->dataPreProcessor = new DataPreProcessor();
        QFutureWatcher<void> *watcher = new QFutureWatcher<void>();
        connect(watcher, SIGNAL(finished()), this, SLOT(dataLoaded()));
        watcher->setFuture(QtConcurrent::run(fileToDataReader, this->dataReader, splitted[splitted.size() - 1], this->dataPreProcessor));

    }
}

/**
 * @brief Funkce volána po dokončení načítání statických dat.
 *
 * Vypína progress bar, povoluje další funkce programu a nastavuje statistiky.
 */
void MainWindow::dataLoaded()
{
    ui->toolButtonOpenFile->setEnabled(true);

    this->disableProgressBar();

    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::StaticRender, true);
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::ProcessSubmap, true);
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::Settings, true);

    this->map->setAllAreas(this->dataPreProcessor->areas);
    this->mapRender->setAllAreas(this->dataPreProcessor->areas);

    this->statistics->setAreaCount(this->dataPreProcessor->areaCount);
    this->statistics->setAverQuality(this->dataPreProcessor->quality);
}

/**
 * @brief Funkce, která je volána při požadavku překreslení statistik.
 */
void MainWindow::reloadStatistics()
{
    if(ui->tabWidgetRightCol->currentIndex() == RightColumnTabs::StatisticsTab) {
        QStandardItemModel *model = new QStandardItemModel(9, 2);
        model->setItem(0, 0, new QStandardItem("Area count:"));
        model->setItem(0, 1, new QStandardItem(QString::number(this->statistics->areaCount())));
        model->setItem(1, 0, new QStandardItem("Quality:"));
        model->setItem(1, 1, new QStandardItem(QString::number(this->statistics->averQuality())));

        model->setItem(2, 0, new QStandardItem("Zoom:"));
        model->setItem(2, 1, new QStandardItem(QString::number(this->statistics->zoom())));
        model->setItem(3, 0, new QStandardItem("Translate:"));
        model->setItem(3, 1, new QStandardItem(QString::number(this->statistics->translateX()) + "x" + QString::number(this->statistics->translateY())));

        if(ui->stackedWidget->currentIndex() == StackedWidget::Map) {
            model->setItem(2, 1, new QStandardItem(QString::number(this->statisticsMap->zoom())));
            model->setItem(3, 1, new QStandardItem(QString::number(this->statisticsMap->translateX()) + "x" + QString::number(this->statisticsMap->translateY())));
        }


        model->setItem(4, 0, new QStandardItem("Reg Index:"));
        model->setItem(4, 1, new QStandardItem(QString::number(this->statistics->staticIndex()) + " to " + QString::number(this->statistics->areaIndex())));
        if(this->dataPreProcessor->areaCount > 0) {
            model->setItem(5, 0, new QStandardItem("Transformable:"));
            QString trans = this->statistics->couldTransform() ? "true" : "false";
            model->setItem(5, 1, new QStandardItem(trans));
            model->setItem(6, 0, new QStandardItem("Sigma:"));
            model->setItem(6, 1, new QStandardItem(QString::number(this->statistics->sigma())));
            model->setItem(7, 0, new QStandardItem("Next Index:"));
            model->setItem(7, 1, new QStandardItem(QString::number(this->statistics->nextStaticIndex()) + " to " + QString::number(this->statistics->nextAreaIndex())));
        }

        if(this->map != NULL) {
            model->setItem(8, 0, new QStandardItem("Submaps:"));
            model->setItem(8, 1, new QStandardItem(QString::number(this->statistics->submapsCount())));
        }

        ui->tableViewStatistics->setModel(model);
        this->setupTableView(ui->tableViewStatistics);
    }
}

/**
 * @brief Překreslení statistik se submapami.
 */
void MainWindow::reloadSubmapTab()
{
    if(ui->tabWidgetRightCol->currentIndex() == RightColumnTabs::SubmapsTab && this->map != NULL) {
        int submapCount = this->map->getSubmapsCount();
        QStandardItemModel *model = new QStandardItemModel(3*submapCount, 2);
        for(int i = 0, h = 0; i < submapCount; i++) {
            int rows = this->map->getPointsCount(i);
            model->setItem(h, 0, new QStandardItem("--Submap--"));
            model->setItem(h, 1, new QStandardItem(QString::number(i + 1)));
            h++;
            model->setItem(h, 0, new QStandardItem("Rows:"));
            model->setItem(h, 1, new QStandardItem(QString::number(rows)));
            h++;
            model->setItem(h, 0, new QStandardItem("***"));
            model->setItem(h, 1, new QStandardItem(""));
            h++;
        }

        ui->tableViewSubmaps->setModel(model);
        this->setupTableView(ui->tableViewSubmaps);
    }
}

/**
 * @brief Překreslení statistik s načtenými mapami.
 */
void MainWindow::reloadLoadedTable()
{
    if(ui->tabWidgetRightCol->currentIndex() == RightColumnTabs::LoadedTab && this->postProc->submapCount != 0) {
        int submapCount = this->postProc->submapCount;
        QStandardItemModel *model = new QStandardItemModel(5*submapCount, 2);
        for(int i = 0, h = 0; i < submapCount; i++) {
            model->setItem(h, 0, new QStandardItem("--Submap--"));
            model->setItem(h, 1, new QStandardItem(QString::number(i + 1)));
            h++;
            model->setItem(h, 0, new QStandardItem("Rows:"));
            model->setItem(h, 1, new QStandardItem(QString::number(this->postProc->loadedMaps[i]->allPoints.n_rows)));
            h++;
            model->setItem(h, 0, new QStandardItem("Robot positions:"));
            model->setItem(h, 1, new QStandardItem(QString::number(this->postProc->loadedMaps[i]->position->position.n_rows)));
            h++;
            model->setItem(h, 0, new QStandardItem("Landmarks:"));
            model->setItem(h, 1, new QStandardItem(QString::number(this->postProc->loadedMaps[i]->landmarksAdded)));
            h++;
            model->setItem(h, 0, new QStandardItem("***"));
            model->setItem(h, 1, new QStandardItem(""));
            h++;
        }

        ui->tableViewLoadedSubmaps->setModel(model);
        this->setupTableView(ui->tableViewLoadedSubmaps);
    }
}

/**
 * @brief Překreslení statistik s informacemi o připojeném LIDARU pomocí UDP.
 */
void MainWindow::reloadLidarTable()
{
    statisticLidar_t *s = this->client->getStatistic();
    QStandardItemModel *model = new QStandardItemModel(8, 2);
    model->setItem(0, 0, new QStandardItem("State:"));
    model->setItem(0, 1, new QStandardItem(this->client->getStateName()));

    model->setItem(1, 0, new QStandardItem("Preprocessing:"));
    model->setItem(1, 1, new QStandardItem(s->preprocessData ? "Yes" : "No"));

    model->setItem(2, 0, new QStandardItem("Address:"));
    model->setItem(2, 1, new QStandardItem(s->address.toString()));

    model->setItem(3, 0, new QStandardItem("Client port:"));
    model->setItem(3, 1, new QStandardItem(QString::number(s->clientPort)));

    model->setItem(4, 0, new QStandardItem("Server port:"));
    model->setItem(4, 1, new QStandardItem(QString::number(s->serverPort)));

    model->setItem(5, 0, new QStandardItem("Path:"));
    model->setItem(5, 1, new QStandardItem(s->path));

    model->setItem(6, 0, new QStandardItem("Sample:"));
    model->setItem(6, 1, new QStandardItem(QString::number(s->sample)));

    model->setItem(7, 0, new QStandardItem("Rate:"));
    model->setItem(7, 1, new QStandardItem(QString::number(s->rate)));

    ui->tableViewLidar->setModel(model);
    this->setupTableView(ui->tableViewLidar);
}

/**
 * @brief Funkce volána při požadavku překreslení jednoho měření.
 */
void MainWindow::reloadStaticRender()
{
    if(this->activeStaticRender == 0) {
        ui->renderAreaScan->setData(this->dataPreProcessor->areas[this->activeStaticRender], true);
    } else {
        ui->renderAreaScan->setData(this->dataPreProcessor->areas[this->activeStaticRender], false);
    }

    this->landmarkExtractionRun();

    ui->renderAreaScan->update();
}

/**
 * @brief Při kliku na tlačítko Reload.
 *
 * Volá funkce allReset() a reloadStaticRender().
 */
void MainWindow::on_toolButton_clicked()
{
    this->allReset();
    this->reloadStaticRender();

    ui->toolButtonNext->setEnabled(true);
}

/**
 * @brief Resetuje hodnoty vykreslených měření.
 */
void MainWindow::allReset()
{
    this->activeStaticRender = 0;
    this->areaIndex = 0;
    //this->preAreaIndex = 0;
    this->correctionCount = 0;
    this->forbiddenIndexes = QVector<int>(100);
}

/**
 * @brief Ukončení aplikace.
 */
void MainWindow::on_toolButtonClose_clicked()
{
    exit(0);
}

/**
 * @brief Zobrazí další měření nebo při registraci spustí vlákna pro hledání transformací.
 */
void MainWindow::nextScan()
{
    if(this->activated && this->completedFirstThreat && this->completedSecondThreat && this->areaIndex < this->dataPreProcessor->areaCount) {
        if(!this->applyRegistration) {
            if(this->dataPreProcessor->areaCount > this->activeStaticRender) {
                ui->renderAreaScan->setData(this->dataPreProcessor->areas[this->activeStaticRender], false);
                this->activeStaticRender++;
                this->nextScan();
                this->statistics->setAreaCount(this->dataPreProcessor->areaCount);
            }
        } else {
            this->completedFirstThreat = false;
            this->completedSecondThreat = false;
            this->registrationStep();
            this->activeStaticRender++;
            this->areaIndex++;
            this->registrationStep();
        }
        ui->renderAreaScan->update();
    }
}

/**
 * @brief Funkce spouští registraci na mračno bodů v jednom vlákně.
 *
 * Po dokončení je volána funkce registrationStepCompleted().
 */
void MainWindow::registrationStep()
{
    this->changeEnableStaticControls(false);

    this->statistics->setAreaCount(this->dataPreProcessor->areaCount);

    this->areaChanged = new area_t();
    this->areaIndex = this->areaIndex == 0 ? this->activeStaticRender + 1 : this->areaIndex;

    if(areaIndex < this->dataPreProcessor->areaCount) {
        QFutureWatcher<void> *watcher = new QFutureWatcher<void>();
        connect(watcher, SIGNAL(finished()), this, SLOT(registrationStepCompleted()));
        watcher->setFuture(QtConcurrent::run(twoPointSetsRegistrationOneStep, this->dataPreProcessor, this->statistics, this->activeStaticRender, this->areaIndex, this->areaChanged));

        // run without thread
        //twoPointSetsRegistrationOneStep(this->dataPreProcessor, this->statistics, this->activeStaticRender, this->areaIndex, this->areaChanged);
    } else {
        this->completedFirstThreat = true;
        this->completedSecondThreat = true;
    }
}

/**
 * @brief Také spouští registraci ve vlastním vlákně, ale jedná se o korekční krok.
 *
 * Tento krok se spouští při nalezení oblastí, které se nepodařilo zarovnat. Po dokončení se volá funkce correctionStepCompleted().
 *
 * @param what - Index oblasti na kterou se zarovnává.
 * @param with - Index oblasti která se zarovnává.
 */
void MainWindow::correctionStep(int what, int with)
{
    if(with < this->dataPreProcessor->areaCount) {
        //qDebug() << "start algorithm: " << what << ", " << with;
        QFutureWatcher<void> *watcher = new QFutureWatcher<void>();
        connect(watcher, SIGNAL(finished()), this, SLOT(correctionStepCompleted()));
        watcher->setFuture(QtConcurrent::run(twoPointSetsRegistrationOneStep, this->dataPreProcessor, this->statistics, what, with, this->areaChanged));
    }
}

/**
 * @brief Korekční krok, teď nepoužíván. Pouze pro testování.
 *
 * Zkouší zarovnat posledních 360 bodů s 360 body už v mapě (začínající na indexu start).
 *
 * @param start - Index počátku oblasti k zarovnání.
 */
void MainWindow::mapCorrectionStep(int start)
{
    arma::mat *allPoints = &this->map->getSubmap(this->map->getSubmapsCount() - 1)->allPoints;
    arma::mat first = allPoints->submat(start - 360, 0, start-1, 1);
    arma::mat second = allPoints->submat(allPoints->n_rows - 360, 0, allPoints->n_rows-1, 1);
    arma::mat result;

    PointSetRegistration *reg = new PointSetRegistration();
    reg->setMatrices(first, second);
    reg->start(result);
}

/**
 * @brief Pokud je dokončen korekční krok.
 *
 * Korekční krok viz. correctionStep(int what, int with). Volá funkci checkRegistrationResults().
 */
void MainWindow::correctionStepCompleted()
{
    this->correctionRunning = false;
    this->checkRegistrationResults();
}

/**
 * @brief Pokud je dokončeno zarovnávání v některém vlákně.
 *
 * Zarovnávání viz. egistrationStep(). Pokud je registrace pořád zapnutá, spouští další krok k zarovnání.
 */
void MainWindow::registrationStepCompleted()
{
    this->changeEnableStaticControls(true);

    this->activeStaticRender++;
    this->areaIndex++;

    if(this->activated) {
        this->registrationStep();
    } else {
        if(!this->completedFirstThreat) {
            this->completedFirstThreat = true;
        } else if(!this->completedSecondThreat) {
            this->completedSecondThreat = true;
        }
    }
}

/**
 * @brief Kontrola oblastí, v kterých se hledata transformace.
 *
 * Tato funkce se spouští ve třetím vlákně a to skládá výslednou mapu. Pokud narazí na oblasti, které nelze zarovnat spustí korekční krok.
 */
void MainWindow::checkRegistrationResults()
{
    if(!this->correctionRunning && this->applyRegistration && this->areaIndex > this->connectedIndex + 1 && this->dataPreProcessor->areas != NULL
            && this->activated) {
        bool connect = false;

        if(dataPreProcessor->areas[this->connectedIndex]->transformable) {
            this->map->setAllAreas(this->dataPreProcessor->areas);
            this->mapRender->setAllAreas(this->dataPreProcessor->areas);
            connectedIndex = connectedWith - 1;
            correctionCount = 0;
            this->map->addAreaToActualSubmap(this->connectedIndex);
            this->mapRender->addAreaToActualSubmap(this->connectedIndex);
        } else {
            this->connectedIndex--;
            connect = true;
        }

        this->connectedWith++;
        this->connectedIndex++;

        if(connectedWith - connectedIndex > 3) {
            this->map->createNewSubmap();
            this->mapRender->createNewSubmap();
            this->connectedIndex = this->connectedWith;
            this->connectedWith++;

        }

        if(connect && !correctionRunning) {
            this->correctionRunning = true;
            this->correctionStep(connectedIndex, connectedWith);
        } else {
            this->correctionRunning = false;
        }

        // check robot position
        /*int ind = this->map->checkRobotPosition();
        if(ind != -1) {
            qDebug() << "close point index: " << ind;
            //this->mapCorrectionStep(ind);
        }*/

        this->statistics->setIndexes(connectedIndex, connectedWith);
    }
}

/**
 * @brief Spuštění hledání důležitých objektů ve vlastním vlákně.
 *
 * Po dokončení ihned volá překreslení v renderArea.
 */
void MainWindow::landmarkExtractionRun()
{
    if(this->landmarkExtraction) {
        QFutureWatcher<void> *watcher = new QFutureWatcher<void>();
        connect(watcher, SIGNAL(finished()), ui->renderAreaScan, SLOT(update()));
        watcher->setFuture(QtConcurrent::run(landmarkExtractionThread, this->dataPreProcessor, this->activeStaticRender, true, true, this->activeStaticRender - 1));
    }
}

/**
 * @brief Při kliku na tlačítko next, se buď načte další jedno měření nebo se spustí další krok zarovnávání.
 */
void MainWindow::on_toolButtonNext_clicked()
{
    if(!this->applyRegistration) {
        this->activeStaticRender++;
        this->reloadStaticRender();
    } else {
        ui->toolButtonStart->setEnabled(true);
        this->registrationStep();
        this->activeStaticRender++;
        this->areaIndex++;
        this->registrationStep();
    }
}

/**
 * @brief Zapnutí automatické registrace.
 */
void MainWindow::on_toolButtonStart_clicked()
{
    this->activated = true;
}

/**
 * @brief Vypnutí automatické registrace.
 */
void MainWindow::on_toolButtonStop_clicked()
{
    this->activated = false;
}

/**
 * @brief Přepnutí do zarovnávacího režimu.
 */
void MainWindow::on_toolButtonRegistration_clicked()
{
    this->applyRegistration = !this->applyRegistration;

    if(this->showLandmark && this->applyRegistration) {
        this->on_toolButtonShowLandmark_clicked();
    }

    ui->toolButtonRegistration->setChecked(this->applyRegistration);
}

/**
 * @brief Mění dostupnost tlačítek v záložce statického vykreslování.
 *
 * Mění tlačítka Reload, Next, Start.
 *
 * @param enabled - Jestli jsou tlačítka dostupná.
 */
void MainWindow::changeEnableStaticControls(bool enabled)
{
    ui->toolButton->setEnabled(enabled);
    ui->toolButtonNext->setEnabled(enabled);
    ui->toolButtonStart->setEnabled(enabled);
}

/**
 * @brief Přepnutí do režimu zobrazení důležitých objektů v aktuálním měření.
 */
void MainWindow::on_toolButtonShowLandmark_clicked()
{
    this->showLandmark = !this->showLandmark;

    if(this->showLandmark && this->applyRegistration) {
        this->on_toolButtonRegistration_clicked();
    }

    if(this->showLandmark) {
        ui->toolButtonStart->setEnabled(false);
    }

    ui->toolButtonShowLandmark->setChecked(this->showLandmark);
    ui->renderAreaScan->setShowCorners(this->showLandmark);
    ui->renderAreaScan->setShowLines(this->showLandmark);
    ui->renderAreaScan->setShowIntersections(this->showLandmark);
}

/**
 * @brief Nastavení statistik pro zobrazení (translate, zoom).
 */
void MainWindow::renderStatistics()
{
    if(ui->stackedWidget->currentIndex() == StackedWidget::ActualScan) {
        this->statistics->setZoom(ui->renderAreaScan->getZoom());
        this->statistics->setTranslate(ui->renderAreaScan->getTranslated());
    } else if(ui->stackedWidget->currentIndex() == StackedWidget::Map) {
        this->statisticsMap->setZoom(ui->renderAreaMap->getZoom());
        this->statisticsMap->setTranslate(ui->renderAreaMap->getTranslated());
    }
}

/**
 * @brief Přepnutí režimu zobrazení na zobrazení aktuálního měření.
 */
void MainWindow::on_toolButtonActualScan_clicked()
{
    ui->toolButtonActualScan->setChecked(true);
    ui->toolButtonMap->setChecked(false);
    ui->toolButtonRealActual->setChecked(true);
    ui->toolButtonRealMap->setChecked(false);
    ui->stackedWidget->setCurrentIndex(StackedWidget::ActualScan);
}

/**
 * @brief Přepnutí režimu zobrazení na zobrazení mapy.
 */
void MainWindow::on_toolButtonMap_clicked()
{
    ui->toolButtonActualScan->setChecked(false);
    ui->toolButtonMap->setChecked(true);
    ui->toolButtonRealActual->setChecked(false);
    ui->toolButtonRealMap->setChecked(true);
    ui->stackedWidget->setCurrentIndex(StackedWidget::Map);
}

/**
 * @brief Vykreslení mapy při každé změně.
 */
void MainWindow::updateMapRender()
{
    if(ui->stackedWidget->currentIndex() == StackedWidget::Map && renderSubmap && !obDraw) {
        int submap = this->mapRender->getSubmapsCount() - 1;
        ui->renderAreaMap->setData(this->mapRender->getSubmatPoints(submap), this->mapRender->getPointsCount(submap), false);
        ui->renderAreaMap->setRobotPosition(this->mapRender->getPosition(submap));
        ui->renderAreaMap->setLandmarks(this->map->getSubmap(submap)->landmarksPosition, this->map->getSubmap(submap)->landmarks, this->map->getSubmap(submap)->landmarksAdded);
        ui->renderAreaMap->update();
    }
    // Vykreslovat co druhé zarovnání
    //this->obDraw = !this->obDraw;
}

/**
 * @brief Přibližení vykreslení.
 */
void MainWindow::zoomIn()
{
    if(ui->stackedWidget->currentIndex() == StackedWidget::ActualScan) {
        ui->renderAreaScan->zoomIn();
    } else if(ui->stackedWidget->currentIndex() == StackedWidget::Map) {
        ui->renderAreaMap->zoomIn();
    }
}

/**
 * @brief Oddálení vykreslení.
 */
void MainWindow::zoomOut()
{
    if(ui->stackedWidget->currentIndex() == StackedWidget::ActualScan) {
        ui->renderAreaScan->zoomOut();
    } else if(ui->stackedWidget->currentIndex() == StackedWidget::Map) {
        ui->renderAreaMap->zoomOut();
    }
}

/**
 * @brief Vykreslení mapy při kliku na položku v jejich seznamu.
 * @param index - Index řádku na který se kliklo.
 */
void MainWindow::onSubmapClicked(QModelIndex index)
{
    this->on_toolButtonMap_clicked();
    int count = this->mapRender->getSubmapsCount();
    this->renderedSubmap = index.row()/3;
    if(renderedSubmap != count - 1) {
        this->renderSubmap = false;
    } else {
        this->renderSubmap = true;
    }
    ui->renderAreaMap->setData(this->mapRender->getSubmatPoints(renderedSubmap), this->mapRender->getPointsCount(renderedSubmap), false);
    ui->renderAreaMap->setRobotPosition(this->mapRender->getPosition(renderedSubmap));
    ui->renderAreaMap->setLandmarks(this->map->getSubmap(renderedSubmap)->landmarksPosition, this->map->getSubmap(renderedSubmap)->landmarks, this->map->getSubmap(renderedSubmap)->landmarksAdded);
    ui->renderAreaMap->update();
}

void MainWindow::checkMainMenuEnabled()
{
    if(ui->mainMenuTabs->currentIndex() == MainMenuTabs::ProcessSubmap) {
    }
}

/**
 * @brief Při kliku na nastaveni LIDARU, přepnutí hlavní části okna.
 */
void MainWindow::on_toolButtonLidarSettings_clicked()
{
    ui->stackedWidget->setCurrentIndex(StackedWidget::LidarSettings);
}

/**
 * @brief Při kliku na tlačítko pro připojení k serveru.
 */
void MainWindow::on_pushButtonConnect_clicked()
{
    ui->pushButtonConnect->setEnabled(false);
    bool ok, ok2;
    quint16 serverPort = ui->lineEditServerPort->text().toUInt(&ok);
    quint16 clientPort = ui->lineEditClientPort->text().toUInt(&ok2);
    QString address = ui->lineEditIPAddress->text();
    if(address.length() > 0 && ok && ok2) {
        this->client->tryConnect(address, serverPort, clientPort);
    }
    if(this->client->getState() == UdpClient::ClientState::NotConnected) {
        ui->pushButtonConnect->setEnabled(true);
    }

    ui->tabWidgetRightCol->setCurrentIndex(RightColumnTabs::LidarTab);
}

/**
 * @brief Po každé změně stavu UDP klienta, vypíše do textového pole informace o stavu nebo seznam modulů.
 */
void MainWindow::clientChangeState()
{
    ui->mainMenuTabs->setTabEnabled(MainMenuTabs::RealTimeRender, false);
    if(this->client->getState() == UdpClient::ClientState::ConnectedToServer) {
        ui->stackedWidgetLidar->setCurrentIndex(StackedWidgetLidar::Connected);
        ui->textEditTerminal->append("Succesfully connected to server.");
        QStringList modules = this->client->getListOfModules();
        ui->textEditTerminal->append("List of modules: ");
        for(int i = 0; i < modules.length(); i++) {
            ui->textEditTerminal->append("   " + modules.at(i));
        }
    } else if(this->client->getState() == UdpClient::ClientState::ConnectedToLidar || this->client->getState() == UdpClient::ClientState::ReceivingData) {
        ui->stackedWidgetLidar->setCurrentIndex(StackedWidgetLidar::Actions);
        if(this->client->getState() == UdpClient::ClientState::ReceivingData) {
            ui->textEditTerminal->append("Receiving data from server.");
            ui->mainMenuTabs->setTabEnabled(MainMenuTabs::RealTimeRender, true);
        } else {
            ui->textEditTerminal->append("Connected to lidar.");
        }
    } else {
        ui->textEditTerminal->append("Disconnected from server.");
        ui->stackedWidgetLidar->setCurrentIndex(StackedWidgetLidar::NotConnected);
    }
}

/**
 * @brief Při kliku na tlačítko pro připojení k serveru.
 */
void MainWindow::on_pushButtonConnectLidar_clicked()
{
    if(ui->lineEditPath->text().length() > 0) {
        this->client->connectToLidar(ui->lineEditPath->text());
    }
}

/**
 * @brief Při kliku na tlačítko Grab (aby server posílal data do aplikace).
 */
void MainWindow::on_pushButtonGrab_clicked()
{
    bool ok1, ok2;
    int sample = ui->lineEditGrab->text().toInt(&ok1);
    int rate = ui->lineEditRate->text().toInt(&ok2);
    if(ok1 && ok2) {
        this->client->startGrabbing(sample, rate);
        this->activated = true;
        this->completedFirstThreat = true;
        this->completedSecondThreat = true;
    }
}

/**
 * @brief Při kliku na tlačítko Stop (zastavení posílání dat ze serveru).
 */
void MainWindow::on_pushButtonStop_clicked()
{
    this->client->stopGrabbing();
    this->activated = false;
}

/**
 * @brief Při kliku na tlačítko StateInfo, požadavek na odeslání informací o stavu ze serveru.
 */
void MainWindow::on_pushButtonStateInfo_clicked()
{
    this->client->stateMessage();
}

/**
 * @brief Při kliku na smazání submapy.
 */
void MainWindow::on_toolButtonRemoveSubmap_clicked()
{
    if(!this->renderSubmap) {
        this->map->removeSubmap(this->renderedSubmap);
        this->mapRender->removeSubmap(this->renderedSubmap);
        this->reloadSubmapTab();
    }
}

/**
 * @brief Změna dostupnosti tlačítek pro záložku submap.
 *
 * Změna pro tlačítka Save, Load, Remove.
 *
 * @param enabled - Jestli jsou dostupná.
 */
void MainWindow::changeEnabledSubmapControls(bool enabled)
{
    ui->toolButtonSaveSubmap->setEnabled(enabled);
    ui->toolButtonLoadSubmap->setEnabled(enabled);
    ui->toolButtonRemoveSubmap->setEnabled(enabled);
}

/**
 * @brief Po kliku na tlačítko pro uložení submapy.
 */
void MainWindow::on_toolButtonSaveSubmap_clicked()
{
    if(!this->activated) {
        this->changeEnabledSubmapControls(false);
        this->postProc->saveSubmap(this->mapRender->getSubmap(this->renderedSubmap), this->map->getSubmap(this->renderedSubmap));
        this->changeEnabledSubmapControls(true);
    }
}

/**
 * @brief Po kliku na tlačítko pro načtení mapy.
 */
void MainWindow::on_toolButtonLoadSubmap_clicked()
{
    QString path = QFileDialog::getOpenFileName();
    if(!path.isEmpty()) {
        QStringList splitted = path.split("/");
        ui->toolButtonLoadSubmap->setEnabled(false);

        QFutureWatcher<void> *watcher = new QFutureWatcher<void>();
        connect(watcher, SIGNAL(finished()), this, SLOT(submapLoaded()));
        watcher->setFuture(QtConcurrent::run(loadSubmap, this->postProc, splitted[splitted.size() - 1]));

        this->enableProgressBar();
    }
}

/**
 * @brief Volá se po načtení submapy.
 */
void MainWindow::submapLoaded()
{
    this->disableProgressBar();
    ui->toolButtonLoadSubmap->setEnabled(true);
    ui->tabWidgetRightCol->setCurrentIndex(RightColumnTabs::LoadedTab);
}

/**
 * @brief Po kliku na Start v realTimRender, spustí registraci na příchozí data.
 */
void MainWindow::on_toolButtonRealTimeStart_clicked()
{
    this->activated = true;
    this->applyRegistration = true;
    this->completedFirstThreat = true;
    this->completedSecondThreat = true;
    this->map->setAllAreas(this->dataPreProcessor->areas);
    this->mapRender->setAllAreas(this->dataPreProcessor->areas);

    this->areaIndex = this->activeStaticRender;
    this->activeStaticRender -= 1;

    this->connectedIndex = this->activeStaticRender-1;
    this->connectedWith = this->areaIndex-1;
}

/**
 * @brief Po kliku na Stop v realTimeRender. Zastaví registraci na příchozí data.
 */
void MainWindow::on_toolButtonRealTimeStop_clicked()
{
    this->activated = false;
    this->applyRegistration = false;
}

/**
 * @brief Po kliku na položku v seznamu submap (vykreslí danou submapu).
 * @param index - Index řádku, na který se kliklo.
 */
void MainWindow::on_tableViewLoadedSubmaps_clicked(const QModelIndex &index)
{
    this->on_toolButtonMap_clicked();
    int submap = index.row()/5;
    int count = this->postProc->loadedMaps[submap]->allPoints.n_rows;
    QPointF *points = new QPointF[count];
    for(int i = 0; i < count; i++) {
        points[i].setX(this->postProc->loadedMaps[submap]->allPoints.at(i, 0));
        points[i].setY(this->postProc->loadedMaps[submap]->allPoints.at(i, 1));
    }

    ui->renderAreaMap->setData(points, count, true);
    ui->renderAreaMap->setRobotPosition(this->postProc->loadedMaps[submap]->position);
    ui->renderAreaMap->setLandmarks(this->postProc->loadedMaps[submap]->landmarksPosition, this->postProc->loadedMaps[submap]->landmarks, this->postProc->loadedMaps[submap]->landmarksCount);
    ui->renderAreaMap->update();
}
