#include "mapclass.h"

#include <QDebug>
#include <qmath.h>

MapClass::MapClass(area_t **allAreas, Statistics *statistics, bool main) : QObject()
{
    this->submapsCount = 1;
    this->actualSubmap = 0;
    this->maxDistance = 10000;

    this->main = main;

    this->submaps = new submap_t*[this->submapsCount];
    this->submaps[0] = new submap_t();

    this->allAreas = allAreas;
    this->statistics = statistics;
    this->statistics->setSubmapsCount(this->submapsCount);
}

/**
 * @brief Funkce k nastavení směrníku na všechny oblasti.
 * @param allAreas - Směrník (pointer) na všechny oblasti.
 */
void MapClass::setAllAreas(area_t **allAreas)
{
    this->allAreas = allAreas;
}

/**
 * @brief Nastavení maximální vzdálenosti bodů, které budou přidány do mapy, používano v addAreaToActualSubmap().
 * Pokud je bod dál od skeneru, než je maxDistance, tak není přidán do mapy.
 * @param maxD - Maximální vzdálenost bodu.
 */
void MapClass::setMaxDistance(int maxD)
{
    this->maxDistance = maxD;
}

/**
 * @brief Funkce k přidávání oblastí area_t do mapy submap_t.
 * Celá aktuální submapa je transformována a následně je připojena nová oblast. Pokud submapa obsahuje víc jak 100 000 bodů, vytvoří se nová.
 * @param areaIndex - Index oblasti, která bude přidána (pro správnou funkci, oblast musí mít po registraci a musí být transformovatelná).
 * @return Počet bodů v mapě bez aktuální oblasti.
 */
int MapClass::addAreaToActualSubmap(int areaIndex)
{
    this->submaps[actualSubmap]->position->addNext(allAreas[areaIndex], this->submaps[actualSubmap]->allPoints.n_rows);
    int areaPointCount = 0;
    mat areaPoints;
    // for speed it up, only one "if" statement istead of checking it on all point.
    if(maxDistance == 10000) {
        areaPointCount = allAreas[areaIndex]->pointCount;
        areaPoints = zeros<mat>(areaPointCount, 2);

        for(int i = 0, h = 0; i < allAreas[areaIndex]->pointCount; i++) {
            areaPoints(h, 0) = allAreas[areaIndex]->points[i].x();
            areaPoints(h, 1) = allAreas[areaIndex]->points[i].y();
            h++;
        }
    } else {
        for(int i = 0; i < allAreas[areaIndex]->pointCount; i++) {
            if(allAreas[areaIndex]->points[i].x() < maxDistance && allAreas[areaIndex]->points[i].y() < maxDistance) {
                areaPointCount++;
            }
        }
        areaPoints = zeros<mat>(areaPointCount, 2);
        for(int i = 0, h = 0; i < allAreas[areaIndex]->pointCount; i++) {
            if(allAreas[areaIndex]->points[i].x() < maxDistance && allAreas[areaIndex]->points[i].y() < maxDistance) {
                areaPoints(h, 0) = allAreas[areaIndex]->points[i].x();
                areaPoints(h, 1) = allAreas[areaIndex]->points[i].y();
                h++;
            }
        }
    }

    // transforming submap and adding new area
    if(submaps[actualSubmap]->firstAreaIndex == -1) {
        submaps[actualSubmap]->firstAreaIndex = areaIndex;
        submaps[actualSubmap]->allPoints = zeros<mat>(areaPointCount, 2);
    } else {
        submaps[actualSubmap]->allPoints.resize(submaps[actualSubmap]->allPoints.n_rows + areaPointCount, 2);
        submaps[actualSubmap]->absoluteRotation *= allAreas[areaIndex]->rotationMatrix;
        submaps[actualSubmap]->absoluteTranslation += allAreas[areaIndex]->translationVector;
        submaps[actualSubmap]->allPoints = submaps[actualSubmap]->allPoints * allAreas[areaIndex]->rotationMatrix -
                repmat(allAreas[areaIndex]->translationVector.t(), submaps[actualSubmap]->allPoints.n_rows, 1);
        if(main) {
            submaps[actualSubmap]->landmarksPosition = submaps[actualSubmap]->landmarksPosition * allAreas[areaIndex]->rotationMatrix -
                    repmat(allAreas[areaIndex]->translationVector.t(), submaps[actualSubmap]->landmarksPosition.n_rows, 1);
        }
    }
    submaps[actualSubmap]->lastAreaIndex = areaIndex;

    int rowsCount = submaps[actualSubmap]->allPoints.n_rows;
    submaps[actualSubmap]->allPoints.submat(rowsCount - areaPointCount, 0, rowsCount - 1, 1) = areaPoints;

    if(main) {
        this->addLandmarksToSubmap(areaIndex);
    }

    //qDebug() << "::MAP::" << " rows=" << submaps[actualSubmap]->allPoints.n_rows;


    if(submaps[actualSubmap]->allPoints.n_rows > 100000) {
        this->createNewSubmap();
    }

    emit this->updated();
    emit this->submapChanged();

    return rowsCount - areaPointCount;
}

/**
 * @brief Přidávání důležitých objektů z oblasti area_t do celé mapy submap_t.
 * Pro každý nový důležitý objekt jsou kontrolovány všechny dosud přidané (jestli už byl přidán v minulosti). Pokud se objeví, že objekt už byl přidán, inkrementuje se jeho důvěryhodnost a zpřesní jeho pozice. Zatím jsou přidavány pouze rohy.
 * @param areaIndex - Index oblasti obsahující důležité objekty.
 */
void MapClass::addLandmarksToSubmap(int areaIndex)
{
    if(this->allAreas[areaIndex]->cornerCount > 0) {
        qreal *lx, *ly;
        QPointF *p, add;
        qreal maxDist = 3500, len, lenStart, lenAdded = 5000;
        bool added;
        int index = 0;
        for(int i = 0; i < this->allAreas[areaIndex]->cornerCount; i++) {
            added = false;
            p = &this->allAreas[areaIndex]->corners[i];
            if(p->x() < maxDist && p->x() > -maxDist && p->y() < maxDist && p->y() > -maxDist) {
                // for all landmarks already added is checked theirs position.
                for(int h = 0; h < this->submaps[actualSubmap]->landmarksAdded; h++) {
                    lx = &this->submaps[actualSubmap]->landmarksPosition.at(h, 0);
                    ly = &this->submaps[actualSubmap]->landmarksPosition.at(h, 1);
                    len = qSqrt(qPow(p->x() - *lx, 2) + qPow(p->y() - *ly, 2));
                    lenStart = qSqrt(qPow(p->x(), 2) + qPow(p->y(), 2));
                    // if distance between already mapped landmark and new landmark is lower then distance from [0, 0] devided by 10.
                    if(len < lenStart/10) {
                        added = true;
                        // if we found closer landmark, it is choosen
                        if(lenAdded > len) {
                            index = h;
                            add.setX(*lx);
                            add.setY(*ly);

                        }
                        lenAdded = len;
                    }
                }
                if(!added) {
                    this->submaps[actualSubmap]->addLandmark(p);
                } else {
                    // in case landmark is already added, position is recalculated in proportion 2/3
                    this->submaps[actualSubmap]->landmarks[index]->truthFactor += 1;
                    this->submaps[actualSubmap]->landmarksPosition[index, 0] = (add.x() + p->x() * 2)/3;
                    this->submaps[actualSubmap]->landmarksPosition[index, 1] = (add.y() + p->y() * 2)/3;
                }
            }
        }
    }
}

/**
* @brief Funkce pro tvorbu nové submapy.
*/
void MapClass::createNewSubmap()
{
    if(this->submaps[this->actualSubmap]->firstAreaIndex != -1) {
        submap_t **submapsOld = new submap_t*[this->submapsCount];
        for(int i = 0; i < submapsCount; i++) {
            submapsOld[i] = this->submaps[i];
        }

        this->submapsCount++;
        this->submaps = new submap_t*[this->submapsCount];
        for(int i = 0; i < this->submapsCount-1; i++) {
            this->submaps[i] = submapsOld[i];
        }
        this->submaps[this->submapsCount-1] = new submap_t();

        this->actualSubmap++;
        this->statistics->setSubmapsCount(this->submapsCount);

        emit submapChanged();
    }
}

/**
 * @brief Vrací směrník na body z požadované submapy.
 * @param submap - Index submapy, z které jsou body požadovány.
 * @return - Směrník na pole bodů.
 */
QPointF *MapClass::getSubmatPoints(int submap)
{
    int count = this->submaps[submap]->allPoints.n_rows;
    this->actualSubmatPoints = new QPointF[count];
    for(int i = 0; i < count; i++) {
        actualSubmatPoints[i].setX(this->submaps[submap]->allPoints(i, 0));
        actualSubmatPoints[i].setY(this->submaps[submap]->allPoints(i, 1));
    }
    this->submatPointsCount = count;
    return this->actualSubmatPoints;
}

/**
 * @brief Vrací počet bodů v požadované submapě.
 * @param submap - Index submapy.
 * @return Počet bodů.
 */
int MapClass::getPointsCount(int submap) const
{
    return this->submaps[submap]->allPoints.n_rows;
}


/**
 * @brief Kontrolování pozice robota, jestli se přiblížil na místo, kde už v minulosti byl.
 * Hledá se nejbližší předchozí pozice robota. Pokud je pozice nalezena, vrací index bodu, protože mapa je složená z oblastí, takže návratová hodnota je index prvního bodu této oblasti.
 * @return Index prvního bodu v submapě kde začíná nalezená oblast.
 */
int MapClass::checkRobotPosition()
{
    qreal d, len, minD;
    qreal *x1, *x2, *y1, *y2;
    len = 0;
    minD = 1000;
    QPointF u;
    int lastCheck = 0, stop = 0;
    for(int i = this->submaps[actualSubmap]->position->position.n_rows-2; i >= 0; i--) {
        x1 = &this->submaps[actualSubmap]->position->position(i, 0);
        x2 = &this->submaps[actualSubmap]->position->position(i+1, 0);
        y1 = &this->submaps[actualSubmap]->position->position(i, 1);
        y2 = &this->submaps[actualSubmap]->position->position(i+1, 1);
        if(len < 2000) {
            u = QPointF(*x1 - *x2,
                        *y1 - *y2);
            len += qSqrt(qPow(u.x(), 2) + qPow(u.y(), 2));
        } else {
            d = qAbs(*x2 * *y1 - *y2 * *x1)/qSqrt(qPow(*y2 - *y1, 2) + qPow(*x2 - *x1, 2));

            if(lastCheck) {
                stop = 1;
            }
            if(d < minD && d <= 30) {
                minD = d;
                lastCheck = this->submaps[actualSubmap]->position->firstPoint.at(i);
            }
            if(stop) {
                return lastCheck;
            }
        }
    }
    return -1;
}

/**
 * @brief Přidává důležitý bod do mapy, kontroluje jejich počet a automaticky realokuje pole.
 * @param point - Pozice důležitého bodu.
 */
void submap_t::addLandmark(QPointF *point)
{
    if(this->landmarksCount > this->landmarksAdded) {
        this->landmarksPosition.at(this->landmarksAdded, 0) = point->x();
        this->landmarksPosition.at(this->landmarksAdded, 1) = point->y();
        this->landmarks[landmarksAdded] = new landmark_t();
        this->landmarks[landmarksAdded]->truthFactor = 0;

        //qDebug() << this->landmarksPosition.at(this->landmarksAdded, 0) << this->landmarksPosition.at(this->landmarksAdded, 1);
        this->landmarksAdded++;
    } else {
        this->landmarksCount += 100;
        this->landmarksPosition.resize(this->landmarksCount, 2);
        landmark_t **oldLandmark = this->landmarks;

        this->landmarks = new landmark_t*[this->landmarksCount];
        for(int i = 0; i < this->landmarksAdded; i++) {
            this->landmarks[i] = oldLandmark[i];
        }
    }
}

/**
 * @brief Funkce k odstranění požadované submapy. Po odstranění je pole reindexováno.
 * @param submap - Index submapy, která má být odstraněna.
 */
void MapClass::removeSubmap(int submap)
{
    submap_t *toRemove = this->submaps[submap];
    for(int i = submap; i < this->submapsCount; i++) {
        this->submaps[i] = this->submaps[i+1];
    }
    this->submapsCount -= 1;
    this->actualSubmap = this->submapsCount - 1;
    delete toRemove;
}
