#include "pointsetregistration.h"

#include <qmath.h>
#include <QDebug>
#include <stdio.h>
#include <fgt/fgt.hpp>

PointSetRegistration::PointSetRegistration() : QObject()
{
    this->first = new area_t();
    this->first->pointCount = 0;

    this->second = new area_t();
    this->second->pointCount = 0;
}

/**
 * @brief Vložení oblastí, na kterých bude prováděna registrace.
 * @param first - První směrník na oblast.
 * @param second - Druhý směrník na oblast.
 */
void PointSetRegistration::setAreas(area_t *first, area_t *second)
{
    this->inicialization(first, second);
}

/**
 * @brief Vložení referancí na oblasti, reprezentovány maticemi. Na těchto oblastech bude prováděna registrace.
 * @param first - První reference na oblast.
 * @param second - Druhá reference na oblast.
 */
void PointSetRegistration::setMatrices(mat &first, mat &second)
{
    this->inicialization(first, second);
}

/**
 * @brief Inicializace registračního algoritmu.
 *
 * Odřízne se přebytečne body, aby oba mračna měla stejný počet bodů. Překopírování hodnot do matic.
 *
 * @param areaFirst - První oblast.
 * @param areaSecond - Druhá oblast.
 */
void PointSetRegistration::inicialization(area_t *areaFirst, area_t *areaSecond)
{
    this->first = areaFirst;
    this->second = areaSecond;

    int firstCount = first->pointCount;
    int secondCount = second->pointCount;

    if(firstCount > secondCount) {
        firstCount = secondCount;
    } else {
        secondCount = firstCount;
    }

    this->firstSetOfPoints = mat(firstCount, 2);
    for(int i = 0; i < firstCount; i++) {
        this->firstSetOfPoints(i, 0) = first->points[i].x();
        this->firstSetOfPoints(i, 1) = first->points[i].y();
    }

    this->secondSetOfPoints = mat(secondCount, 2);
    for(int i = 0; i < secondCount; i++) {
        this->secondSetOfPoints(i, 0) = second->points[i].x();
        this->secondSetOfPoints(i, 1) = second->points[i].y();
    }

    this->rotationMatrixR = eye<mat>(2, 2);
    this->translationT = zeros<vec>(2);
    this->sigma = 0;
}

/**
 * @brief Inicializace pro oblasti (mračna bodů) zadané maticemi.
 * @param first - První oblast.
 * @param second - Druhá oblast.
 */
void PointSetRegistration::inicialization(mat &first, mat &second)
{
    if(first.n_rows == second.n_rows) {
        this->firstSetOfPoints = first;
        this->secondSetOfPoints = second;
        this->rotationMatrixR = eye<mat>(2, 2);
        this->translationT = zeros<vec>(2);
        this->sigma = 0;
    } else {
        qDebug() << "Matrices must have the same size!";
    }
}


/**
 * @brief Start registračního algoritmu.
 *
 * V první fázi probíhá normalizace matic na bezrozměrné hodnoty. Následně se vypočítá rozptyl sigma a spustí se registrace. Výsledek registrace, matice T se nakonec denormalizuje a vrátí se jako oblast area_t.
 *
 * @param result - Transformovaná (zarovnaná) oblast.
 */
void PointSetRegistration::start(area_t &result)
{
    int firstCount = this->firstSetOfPoints.n_rows;
    int secondCount = this->secondSetOfPoints.n_rows;

    //qDebug() << firstCount << "COUNT" << secondCount;

    normalization_t normal = normalize(this->firstSetOfPoints, this->secondSetOfPoints);

    sigma = 0;
    this->sigma = (this->secondSetOfPoints.n_rows * trace(this->firstSetOfPoints.t() * this->firstSetOfPoints) + this->firstSetOfPoints.n_rows *
                   trace(this->secondSetOfPoints.t() * this->secondSetOfPoints) - as_scalar(2 * sum(this->firstSetOfPoints) * sum(this->secondSetOfPoints).t())) /
            (this->secondSetOfPoints.n_rows * this->firstSetOfPoints.n_rows * this->secondSetOfPoints.n_cols);

    this->corespondenceMatrixP = zeros<mat>(secondCount, firstCount);

    mat T;
    this->startAlgorithm(T);

    this->denormalize(T, normal);

    result.pointCount = 0;
    result.points = new QPointF[T.n_rows];
    for(uword i = 0; i < T.n_rows; i++) {
        result.points[i].setX(T(i, 0));
        result.points[i].setY(T(i, 1));
        result.pointCount = i;
    }

    this->second->rotationMatrix = this->rotationMatrixR;
    this->second->translationVector = normal.scale * this->translationT + normal.xd.t() - this->rotationMatrixR * normal.yd.t();
    this->second->sigma2 = this->sigma;
}

/**
 * @brief Start registračního algoritmu pro oblasti zadané maticemi.
 *
 * Pro více informací viz. PointSetRegistration::start(area_t &result).
 *
 * @param result - Zarovnaná oblast reprezentována maticí.
 */
void PointSetRegistration::start(mat &result)
{
    int firstCount = this->firstSetOfPoints.n_rows;
    int secondCount = this->secondSetOfPoints.n_rows;

    //qDebug() << firstCount << "COUNT" << secondCount;

    normalization_t normal = normalize(this->firstSetOfPoints, this->secondSetOfPoints);

    sigma = 0;
    this->sigma = (this->secondSetOfPoints.n_rows * trace(this->firstSetOfPoints.t() * this->firstSetOfPoints) + this->firstSetOfPoints.n_rows *
                   trace(this->secondSetOfPoints.t() * this->secondSetOfPoints) - as_scalar(2 * sum(this->firstSetOfPoints) * sum(this->secondSetOfPoints).t())) /
            (this->secondSetOfPoints.n_rows * this->firstSetOfPoints.n_rows * this->secondSetOfPoints.n_cols);

    this->corespondenceMatrixP = zeros<mat>(secondCount, firstCount);

    mat T;
    this->startAlgorithm(T);

    this->denormalize(T, normal);

    /*result.pointCount = 0;
    result.points = new QPointF[T.n_rows];
    for(uword i = 0; i < T.n_rows; i++) {
        result.points[i].setX(T(i, 0));
        result.points[i].setY(T(i, 1));
        result.pointCount = i;
    }*/

    result = T;

    qDebug() << "MapCorrectionSigma: " << this->sigma;

    //this->second->rotationMatrix = this->rotationMatrixR;
    //this->second->translationVector = normal.scale * this->translationT + normal.xd.t() - this->rotationMatrixR * normal.yd.t();
    //this->second->sigma2 = this->sigma;
}

/**
 * @brief Samotná registrace mračen bodů.
 *
 * Na začátku probíha inicializace, potom se spouští registrace ve smyčce while s ukončovacími podmínkami. V každém kroce se vypočítá P, A, R, s, t, T.
 *
 * @param T - Výsledné zarovnané mračno bodů.
 */
void PointSetRegistration::startAlgorithm(mat &T)
{
    const uword M = this->secondSetOfPoints.n_rows;
    const uword D = this->secondSetOfPoints.n_cols;

    vec transAll = zeros<vec>(2);

    int iteration = 0;
    double s = 1;
    double w = 0.5;
    double L = 0;
    double L_old, Np;
    double ntol = 1e-5f + 10;

    T = this->secondSetOfPoints;

    vec P1(M), Pt1(M), mu_x(D), mu_y(D), sv;
    mat PX(M, D), A(D, D), PY, U(D, D), V(D, D), C(D, D), S;

    double oldSigma = 0;
    uint sigmaCount = 0;

    vec onesV = ones<vec>(this->firstSetOfPoints.n_rows);

    while(iteration < 100 && (sigma > 4e-5f) && sigmaCount < 10) {
        if(iteration > 30 && sigma > 1e-4) {
            iteration = 100;
        }

        // matrix P
        L_old = L;
        L = this->find_P(this->firstSetOfPoints, T, this->sigma, w, P1, Pt1, PX);

        ntol = std::abs((L - L_old) / L);

        //qDebug() << "rigid iteration: dL " << ntol << ", iter= " << iteration << "scale=" << s << ", sigma2= " << QString::number(this->sigma, 'g', 15);

        // M-step
        Np = sum(Pt1);
        if(Np == 0) {
            Np = 1e-6f;
        }
        mu_x = this->firstSetOfPoints.t() * Pt1 / Np;
        mu_y = this->secondSetOfPoints.t() * P1 / Np;

        A = PX.t() * this->secondSetOfPoints - Np * (mu_x * mu_y.t());

        C = eye<mat>(D, D);
        svd(U, sv, V, A);
        S = diagmat(sv);

        //qDebug() << "rigid iteration: dL " << ntol << ", iter= " << iteration << ", sigma2= " << this->sigma << "s=" << s;

        C(D - 1, D - 1) = det(U * V.t());

        this->rotationMatrixR = U * C * V.t();

        PY = this->secondSetOfPoints - onesV * mu_y.t();

        s = trace(A.t() * this->rotationMatrixR) / trace(PY.t() * diagmat(P1) * PY);

        this->sigma = std::abs(accu(pow(this->firstSetOfPoints, 2) % repmat(Pt1, 1, D)) -
                          Np * as_scalar(mu_x.t() * mu_x) -
                          s * trace(S * C)) /
                (Np * D);

        if(round(oldSigma * 10000000000) == round(sigma * 10000000000)) {
            sigmaCount++;
        } else {
            sigmaCount = 0;
        }
        oldSigma = sigma;

        this->translationT = mu_x - s * this->rotationMatrixR * mu_y;
        transAll += this->translationT;
        T = s * this->secondSetOfPoints * this->rotationMatrixR.t() + repmat(this->translationT.t(), M, 1);
        iteration++;
    }

    this->second->scale = s;

    //qDebug() << "rigid iteration: dL " << ntol << ", iter= " << iteration << ", sigma2= " << this->sigma << ", scale=" << s;
    //qDebug() << "rotation matrix" << this->rotationMatrixR(0, 0) << this->rotationMatrixR(0, 1) << this->rotationMatrixR(1, 0) << this->rotationMatrixR(1, 1);
    //qDebug() << "translation vector" << this->translationT(0) << this->translationT(1);
}

/**
 * @brief Algoritmus k hledání produktů matice P. Pro více info viz. dokumentace k FGT
 * @param X - První mračno bodů.
 * @param Y - Druhé mračno bodů.
 * @param sigma2 - Aktuální rozptyl.
 * @param outliers - Kolik procent bodů jsou tzv. outliers.
 * @param P1 - Výsledný produkt FGT.
 * @param Pt1 - Výsledný produkt FGT.
 * @param PX - Výsledný produkt FGT.
 * @return Míra o kolik se mračno pohlo.
 */
double PointSetRegistration::find_P(const mat &X, const mat &Y, double sigma2, float outliers, vec &P1, vec &Pt1, mat &PX)
{
    P1.zeros();
    Pt1.zeros();
    PX.zeros();

    float epsilon = 1e-4f;

    const uword N = X.n_rows;
    const uword M = Y.n_rows;
    const uword D = Y.n_cols;

    const double h = qSqrt(2 * sigma2); // bandwidth
    const double ndi = (outliers * M * qPow(2 * M_PI * sigma2, 0.5 * D)) / ((1 - outliers) * N);

    vec q = ones<vec>(M);

    fgt::GaussTransformUnqPtr transformY;
    transformY = fgt::choose_gauss_transform(Y, h, epsilon);

    //qDebug() << transformY->get_dimensions() << X.n_rows << q.n_rows;

    vec denomP = transformY->compute(X, q);

    denomP = denomP + ndi;
    Pt1 = 1 - ndi / denomP;
    q = 1 / denomP;

    fgt::GaussTransformUnqPtr transformX;
    transformX = fgt::choose_gauss_transform(X, h, epsilon);
    P1 = transformX->compute(Y, q);

    for(uword i = 0; i < D; i++) {
        vec oneCol = X.col(i);
        q = oneCol / denomP;
        PX.col(i) = transformX->compute(Y, q);
    }

    return -sum(log(denomP)) + D * N * std::log(sigma2) / 2;
}

/**
 * @brief Normalizace na bezrozměrné veličiny, mračen bodů.
 * @param first - První mračno bodů.
 * @param second - Druhé mračno bodů.
 * @return Parametry výsledné normalizace.
 */
normalization_t PointSetRegistration::normalize(mat &first, mat &second)
{
    normalization_t normal;

    const uword N = first.n_rows;
    const uword M = second.n_rows;

    //first.col(2) = first.col(2) * 1.0f;

    normal.xd = mean(first);
    normal.yd = mean(second);

    first = first - repmat(normal.xd, N, 1);
    second = second - repmat(normal.yd, M, 1);

    normal.scale = std::max(scale(first), scale(second));

    first = first / normal.scale;
    second = second / normal.scale;

    return normal;
}

/**
 * @brief Denormalizace mračen bodů.
 * @param result - Mračno bodů k denormalizaci.
 * @param normal - Normalizační parametry.
 */
void PointSetRegistration::denormalize(mat &result, const normalization_t &normal) const
{
    result = result * normal.scale + repmat(normal.xd, result.n_rows, 1);
    //result.col(2) = result.col(2) / 1.0f;
}

/**
 * @brief Výpočet měřítka při normalizaci.
 * @param mat - Mračno bodů.
 * @return Výsledné měřítko.
 */
double PointSetRegistration::scale(const mat &mat) {
    return as_scalar(max(max(mat, 0) - min(mat, 0), 1));
}

/**
 * @brief Velikost vektoru, daného dvěmi body.
 * @param vec1 - První bod (řádek matice).
 * @param vec2 - Druhý bod (řádek matice).
 * @return Výsledná velikost vektoru.
 */
double PointSetRegistration::vectorSize(const vec &vec1, const vec &vec2)
{
    vec u = zeros<vec>(2);
    u(0) = vec1.at(0) - vec2.at(0);
    u(1) = vec1.at(1) - vec2.at(1);
    return qSqrt(qPow(u(0), 2) + qPow(u(1), 2));
}

