#include "renderarea.h"

#include <QPainter>
#include <QDebug>
#include <QKeyEvent>
#include <qmath.h>

/**
 * @brief Inicializace kreslicí plochy, nastavení barev a reset nastavení.
 * @param parent
 */
RenderArea::RenderArea(QWidget *parent) : QWidget(parent)
{
    antialiased = false;
    transformed = false;
    mousePressed = false;
    showCorners = false;
    showLines = false;
    showIntersections = false;

    QPalette pal(palette());
    pal.setColor(QPalette::Background, Qt::black);
    setAutoFillBackground(true);
    setPalette(pal);

    pen.setColor(QColor(Qt::red));
    pen.setWidth(10);

    penSecond.setColor(QColor(Qt::blue));
    penSecond.setWidth(10);

    robotPen.setColor(QColor(Qt::green));
    robotPen.setWidth(8);

    cornerPen.setColor(QColor(Qt::yellow));
    cornerPen.setWidth(65);

    linePen.setColor(QColor(Qt::green));
    linePen.setWidth(8);

    area = new area_t();
    area->pointCount = 0;
    area->points = NULL;

    area_sec = new area_t();
    area_sec->pointCount = 0;
    area_sec->points = NULL;

    position = new robotPosition_t();

    this->resetSettings();

    this->translateArea = QPointF(0, 0);
    this->landmarkCount = 0;
}

void updated();

/**
 * @brief Nastavení pera pro vykreslování bodů.
 * @param pen
 */
void RenderArea::setPen(const QPen &pen)
{
    this->pen = pen;
    update();
}

/**
 * @brief Nastavení brushe pro vybarvování.
 * @param brush
 */
void RenderArea::setBrush(const QBrush &brush)
{
    this->brush = brush;
    update();
}

void RenderArea::setAntialiased(bool antialiased)
{
    this->antialiased = antialiased;
    update();
}

void RenderArea::setTransformed(bool transformed)
{
    this->transformed = transformed;
    update();
}

/**
 * @brief Nastavení zobrazení rohů.
 * @param show - Jestli se mají zobrazovat rohy.
 */
void RenderArea::setShowCorners(bool show)
{
    this->showCorners = show;
    update();
}

/**
 * @brief Nastavení zobrazování hran.
 * @param show - Jestli se mají zobrazovat hrany.
 */
void RenderArea::setShowLines(bool show)
{
    this->showLines = show;
    update();
}

/**
 * @brief Nastavení zobrazování průsečíků hran.
 * @param show - Jestli se mají zobrazovat průsečíky.
 */
void RenderArea::setShowIntersections(bool show)
{
    this->showIntersections = show;
    update();
}

/**
 * @brief Nastavení oblasti, která se má vykreslit.
 * @param area - Oblast k vykreslení.
 * @param reset - Jestli se má resetovat nastavení.
 */
void RenderArea::setData(area_t *area, bool reset)
{
    this->area = area;

    if(reset) {
        this->resetSettings();
    }
}

/**
 * @brief Nastavení dvou oblastí, které se mají vykreslit najednou.
 * @param area - První oblast k vykreslení.
 * @param area_sec - Druhá oblast k vykreslení.
 * @param reset - Jestli se má resetovat nastavení.
 */
void RenderArea::setData(area_t *area, area_t *area_sec, bool reset)
{
    this->area = area;
    this->area_sec = area_sec;

    if(reset) {
        this->resetSettings();
    }
}

/**
 * @brief Nastavení bodů, které se mají vykreslit.
 * @param points - Pole bodů.
 * @param pointCount - Počet bodů.
 * @param reset - Jestli se má resetovat nastavení.
 */
void RenderArea::setData(QPointF *points, int pointCount, bool reset)
{
    this->area->points = points;
    this->area->pointCount = pointCount;

    if(reset) {
        this->resetSettings();
    }
}

/**
 * @brief Nastavení důležitých objektů, které se mají vykreslit.
 * @param landmarkPositions - Pozice důležitých objektů.
 * @param land - Pole objektů, spolu s parametry.
 * @param count - Počet objektů.
 */
void RenderArea::setLandmarks(arma::mat landmarkPositions, landmark_t **land, int count)
{
    int n = count;
    this->landmarks = new QPointF[n];
    this->circle = new qreal[n];
    this->landmarkCount = n;
    for(int i = 0; i < n; i++) {
        //qDebug() << "i: " << landmarkPositions.at(i, 0) << landmarkPositions.at(i, 1);
        landmarks[i].setX(landmarkPositions.at(i, 0));
        landmarks[i].setY(landmarkPositions.at(i, 1));
        this->circle[i] = land[i]->truthFactor * 10;
    }
    this->showCorners = true;
}

/**
 * @brief Nastavení cesty robota, která se má vykreslit.
 * @param position - Pozice robota k vykreslení.
 */
void RenderArea::setRobotPosition(robotPosition_t *position)
{
    this->position = position;
}

/**
 * @brief Tento event se volá při každém překreslení.
 */
void RenderArea::paintEvent(QPaintEvent * /* event */)
{
    this->paintNormal();

    emit updated();
}

/**
 * @brief Vykreslení všech nastavených dat. Vykreslování se provádí mezi painter.save() a painter.restore().
 */
void RenderArea::paintNormal()
{
    QPainter painter(this);
    painter.setPen(pen);
    painter.setBrush(brush);
    if (antialiased) {
        painter.setRenderHint(QPainter::Antialiasing, true);
    }

    painter.save();

    painter.setPen(pen);
    painter.translate(this->translateArea);
    painter.scale(zoom, zoom);
    painter.drawPoints(area->points, area->pointCount);

    if(area->pointCount > 0) {
        painter.setPen(robotPen);
        painter.drawEllipse(QPoint(0, 0), 50, 50);

        if(this->showCorners && this->landmarkCount > 0) {
            painter.setPen(cornerPen);
            painter.drawPoints(this->landmarks, this->landmarkCount);
            painter.setPen(linePen);
            for(int i = 0; i < this->landmarkCount; i++) {
                painter.drawEllipse(this->landmarks[i], this->circle[i], this->circle[i]);
            }
        } else if(this->showCorners) {
            painter.setPen(cornerPen);
            painter.drawPoints(area->corners, area->cornerCount);
        }

        if(this->showLines) {
            painter.setPen(linePen);
            painter.drawLines(area->lines, area->lineCount);
        }

        /*if(this->showIntersections && area->landmarkFinished) {
            painter.setPen(cornerPen);
            int count = 0.5 * area->lineCount * (area->lineCount - 1);
            for(int i = 0; i < count; i++) {
                if(area->intersections[i].angle > 0.1745 && area->intersections[i].angle < 2.967) {
                    painter.drawPoint(area->intersections[i].intersection);
                }
            }
        }*/

        if(this->position->position.n_rows > 1) {
            painter.setPen(linePen);
            for(int i = 0; i < this->position->position.n_rows - 1; i++) {
                QPointF first = QPointF(this->position->position(i, 0), this->position->position(i, 1));
                QPointF second = QPointF(this->position->position(i+1, 0), this->position->position(i+1, 1));
                painter.drawLine(first, second);
            }
        }
    }

    if(area_sec->pointCount > 0) {
        painter.setPen(penSecond);
        painter.drawPoints(area_sec->points, area->pointCount);
    }

    painter.restore();

    painter.setRenderHint(QPainter::Antialiasing, false);
    painter.setPen(palette().dark().color());
    painter.setBrush(Qt::NoBrush);
    painter.drawRect(QRect(0, 0, width() - 1, height() - 1));
}

/**
 * @brief Reset nastavení jako posunutí vykreslovací oblasti a zoomování.
 */
void RenderArea::resetSettings()
{
    translateArea = QPointF(this->width()/2, this->height()/2);
    zoom = 0.06;

    emit this->statistisUpdate();
}

/**
 * @brief Event key press, teď nepoužito.
 * @param event
 */
void RenderArea::keyPressEvent(QKeyEvent *event)
{
    /*if(event->key() == Qt::Key_Left) {
        qDebug() << "leftArrow";
        startDraw.setX(startDraw.x() - 50);
    } else if(event->key() == Qt::Key_Right) {
        startDraw.setX(startDraw.x() + 50);
    } else if(event->key() == Qt::Key_Up) {
        startDraw.setY(startDraw.y() - 50);
    } else if(event->key() == Qt::Key_Down) {
        startDraw.setY(startDraw.y() + 50);
    }
    update();*/
}

/**
 * @brief Event na mouse press.
 * @param event
 */
void RenderArea::mousePressEvent(QMouseEvent *event)
{
    if(event->button() == Qt::LeftButton) {
        this->mousePressed = true;
        this->mousePlace = event->pos();
    }
}

/**
 * @brief Event na mouse move (posunutí vykreslovací oblasti).
 * @param event
 */
void RenderArea::mouseMoveEvent(QMouseEvent *event)
{
    if(this->mousePressed && area->pointCount != 0) {
        this->translateArea.setX(translateArea.x() - (mousePlace.x() - event->pos().x()));
        this->translateArea.setY(translateArea.y() - (mousePlace.y() - event->pos().y()));

        this->mousePlace = event->pos();

        this->update();
        emit updated();
    }
}

/**
 * @brief Event na mouse release.
 */
void RenderArea::mouseReleaseEvent(QMouseEvent *)
{
    this->mousePressed = false;

    emit this->statistisUpdate();
}

/**
 * @brief Vrací hodnotu posunutí vykreslovací oblasti.
 * @return Výsledne posunutí v x a y.
 */
QPointF RenderArea::getTranslated()
{
    QPointF transRelative(this->translateArea.x() - this->width()/2, this->translateArea.y() - this->height()/2);
    return transRelative;
}

/**
 * @brief Vrací hodnotu přibližení.
 * @return Hodnota přibližení.
 */
qreal RenderArea::getZoom()
{
    return this->zoom;
}

/**
 * @brief Přibližení vykreslování.
 */
void RenderArea::zoomIn()
{
    zoom += 0.05;
    this->update();
    emit this->statistisUpdate();
}

/**
 * @brief Oddálení vykreslování.
 */
void RenderArea::zoomOut()
{
    zoom -= 0.05;
    this->update();
    emit this->statistisUpdate();
}

