#include "submappostproc.h"

#include <QDateTime>

#include <QDebug>

SubmapPostProc::SubmapPostProc(QObject *parent) : QObject(parent)
{
    this->submapCount = 0;
    this->pointAdded = 0;
    this->positionAdded = 0;
    this->activeSubmap = -1;
    this->actualState = State::Undefined;
}

/**
 * @brief Ukládání submapy do souboru. Submapa bude uložena do aktuálního adresáře s příponou .smap.
 *
 * V programu se používají dvě submapy, jedna k zarovnávání a hledání důležitých objektů a druhá pouze k vykreslování (zbavena bodů, které leží daleko).
 *
 * @param submap - Submapa k uložení.
 * @param withLandmarks - Submapa s důležitými objekty.
 */
void SubmapPostProc::saveSubmap(submap_t *submap, submap_t *withLandmarks)
{
    QFile file("export-" + QDateTime().currentDateTime().toString("hh-mm-ss_dd-MM-yyyy")+ ".smap");
    file.open(QIODevice::Append);
    file.write("#points\n");
    file.write(QString::number(submap->allPoints.n_rows).toStdString().c_str());
    file.write("\n");
    for(int i = 0; i < submap->allPoints.n_rows; i++) {
        file.write(QString(QString::number(submap->allPoints.at(i, 0)) + "," + QString::number(submap->allPoints.at(i, 1)) + "\n").toStdString().c_str());
    }
    file.write("#robot\n");
    file.write(QString::number(submap->position->position.n_rows).toStdString().c_str());
    file.write("\n");
    for(int i = 0; i < submap->position->position.n_rows; i++) {
        QString text = QString::number(submap->position->position.at(i, 0)) + "," + QString::number(submap->position->position.at(i, 1)) + "\n";
        file.write(text.toStdString().c_str());
    }
    file.write("#landmarks\n");
    file.write(QString::number(withLandmarks->landmarksAdded).toStdString().c_str());
    file.write("\n");
    for(int i = 0; i < withLandmarks->landmarksAdded; i++) {
        QString text = QString::number(withLandmarks->landmarksPosition.at(i, 0)) + "," + QString::number(withLandmarks->landmarksPosition.at(i, 1)) +
                "," + QString::number(withLandmarks->landmarks[i]->truthFactor) + "," + QString::number(withLandmarks->landmarks[i]->type) + "\n";
        file.write(text.toStdString().c_str());
    }
    file.close();
}

/**
 * @brief Nahrávání submapy. Je potřeba aby byla ve stejném adresáři jako aplikace.
 * @param path - Cesta k submapě.
 */
void SubmapPostProc::loadSubmap(const QString path)
{
    this->pointAdded = 0;
    this->positionAdded = 0;

    this->resizeLoadedMaps();
    this->activeSubmap++;

    QFile file(path);
    file.open(QIODevice::ReadOnly);
    while (!file.atEnd()) {
        QByteArray line = file.readLine();
        this->processLine(line, &file);
    }
    file.close();
    emit submapLoaded();
}

/**
 * @brief Čtení jednoho řádku ze souboru.
 * @param line - Jeden řádek.
 * @param file - Soubor.
 */
void SubmapPostProc::processLine(QByteArray &line, QFile *file)
{
    if(line.at(0) == '#') {
        QByteArray lineNext = file->readLine();
        int count = QString(lineNext.data()).replace("\n", "").toInt();
        if(QString(line.data()).compare("#points\n") == 0) {
            this->changeState(State::Points, count);
        } else if(QString(line.data()).compare("#robot\n") == 0) {
            this->changeState(State::Robot, count);
        } else if(QString(line.data()).compare("#landmarks\n") == 0) {
            this->changeState(State::Landmarks, count);
        }
    } else {
        QStringList lineContent = QString(line.data()).replace("\n", "").split(",");
        if(this->actualState == State::Points) {
            this->setPoint(lineContent.at(0).toFloat(), lineContent.at(1).toFloat());
        } else if(this->actualState == State::Robot) {
            this->setRobotPos(lineContent.at(0).toFloat(), lineContent.at(1).toFloat());
        } else if(this->actualState == State::Landmarks) {
            this->setLandmark(lineContent.at(0).toFloat(), lineContent.at(1).toFloat(), lineContent.at(2).toFloat(), lineContent.at(3).toInt());
        }
    }
}

/**
 * @brief Změna stavu načítaných dat.
 *
 * Na začátku je stav nedefinován a pokud se narazí na příznak pro body nastaví se na State::Points atd. Takže nalezené souřadnice se přiřazují do bodů, do té doby než se narazí na další příznak.
 *
 * @param state - Nalezený stav.
 * @param count - Počet prvků v tomto stavu.
 */
void SubmapPostProc::changeState(int state, int count)
{
    if(state == State::Points) {
        this->loadedMaps[this->activeSubmap]->allPoints = zeros<mat>(count, 2);
        this->actualState = State::Points;
    } else if(state == State::Robot) {
        this->loadedMaps[this->activeSubmap]->position->position = zeros<mat>(count, 2);
        this->actualState = State::Robot;
    } else if(state == State::Landmarks) {
        this->loadedMaps[this->activeSubmap]->landmarks = new landmark_t*[count];
        this->loadedMaps[this->activeSubmap]->landmarksAdded = 0;
        this->loadedMaps[this->activeSubmap]->landmarksCount = count;
        this->loadedMaps[this->activeSubmap]->landmarksPosition = zeros<mat>(count, 2);
        this->actualState = State::Landmarks;
    }
}

/**
 * @brief Nastavení bodu podle souřadnic.
 * @param x
 * @param y
 */
void SubmapPostProc::setPoint(qreal x, qreal y)
{
    this->loadedMaps[this->activeSubmap]->allPoints(this->pointAdded, 0) = x;
    this->loadedMaps[this->activeSubmap]->allPoints(this->pointAdded, 1) = y;
    this->pointAdded++;
}

/**
 * @brief Nastavení pozice robota podle souřadnic.
 * @param x
 * @param y
 */
void SubmapPostProc::setRobotPos(qreal x, qreal y)
{
    this->loadedMaps[this->activeSubmap]->position->position(this->positionAdded, 0) = x;
    this->loadedMaps[this->activeSubmap]->position->position(this->positionAdded, 1) = y;
    this->positionAdded++;
}

/**
 * @brief Nastavení pozice a parametru důležitého bodu.
 * @param x
 * @param y
 * @param truth - Důvěryhodnost objektu.
 * @param type - Typ objektu, 0 - roh, 1 - hrana.
 */
void SubmapPostProc::setLandmark(qreal x, qreal y, qreal truth, int type)
{
    int added = this->loadedMaps[this->activeSubmap]->landmarksAdded;
    this->loadedMaps[this->activeSubmap]->landmarks[added] = new landmark_t();
    this->loadedMaps[this->activeSubmap]->landmarks[added]->truthFactor = truth;
    this->loadedMaps[this->activeSubmap]->landmarks[added]->type = type;
    this->loadedMaps[this->activeSubmap]->landmarksPosition(added, 0) = x;
    this->loadedMaps[this->activeSubmap]->landmarksPosition(added, 1) = y;
    this->loadedMaps[this->activeSubmap]->landmarksAdded++;
}

/**
 * @brief Realokace pole, v kterém jsou uložené všechny nahrané mapy.
 */
void SubmapPostProc::resizeLoadedMaps()
{
    submap_t **old = this->loadedMaps;
    this->loadedMaps = new submap_t*[this->submapCount + 1];
    for(int i = 0; i < this->submapCount; i++) {
        this->loadedMaps[i] = new submap_t();
        this->loadedMaps[i] = old[i];
    }
    this->loadedMaps[this->submapCount] = new submap_t();
    this->submapCount += 1;
}

