#include "udpclient.h"

#include <QDebug>

#include <QTimer>
#include <QtEndian>
#include <QFile>

/**
 * @brief Inicializace proměnných a nastavení timeru a zprávy pro heartbeat.
 * @param preProcessor - Třída obsahující naměřené oblasti (tam se také nové budou ukládat).
 * @param parent
 */
UdpClient::UdpClient(DataPreProcessor *preProcessor, QObject *parent) : QObject(parent)
{
    this->state = ClientState::NotConnected;

    this->stat = new statisticLidar_t();
    this->stat->preprocessData = false;

    this->preProcessor = preProcessor;

    QTimer *timer = new QTimer();
    connect(timer, SIGNAL(timeout()), this, SLOT(sendHeartbeat()));
    timer->start(1000);

    this->heartbeatMessage.append('\xff');
    heartbeatMessage.append(rpdev_server::Command::heartbeat);
}

/**
 * @brief Připojení k serveru.
 * @param address - Adresa serveru.
 * @param serverPort - Port serveru.
 * @param clientPort - Port klienta (pro příjem).
 */
void UdpClient::tryConnect(QString address, quint16 serverPort, quint16 clientPort)
{
    if(this->state == ClientState::NotConnected) {
        this->stat->serverPort = serverPort;
        this->stat->clientPort = clientPort;
        this->stat->address = QHostAddress(address);
        this->initSocket();
    }
}

/**
 * @brief Inicializace UDP socketu.
 */
void UdpClient::initSocket()
{
    //qDebug() << "init Socket, client port: " << stat->clientPort << ", server port: " << stat->serverPort;
    udpSocket = new QUdpSocket(this);
    connect(udpSocket, SIGNAL(readyRead()), this, SLOT(readPendingDatagrams()));

    udpSocket->bind(stat->address, stat->clientPort);

    QByteArray msg;
    msg.append('\xff');
    msg.append(rpdev_server::Command::establishConnection);
    msg.append('H');
    msg.append('E');
    msg.append('L');
    msg.append('L');
    msg.append('O');

    udpSocket->writeDatagram(msg, stat->address, stat->serverPort);

    emit this->statisticChange();
}

/**
 * @brief Odeslání datagramu pro připojení k LIDARU.
 * @param path - Cesta k LIDARU.
 */
void UdpClient::connectToLidar(QString path)
{
    this->stat->path = path;

    QByteArray msg;
    msg.append('\x01');
    msg.append(dev_rplidar::Command::Connect);
    msg.append('\x00'); //0
    msg.append('\x01'); //1
    msg.append('\xC2'); //194
    msg.append('\x00'); //0
    msg.append(path.length());
    msg.append(path);

    //qDebug() << msg.size() << path.length();

    this->udpSocket->writeDatagram(msg, this->stat->address, this->stat->serverPort);

    emit this->statisticChange();
}

/**
 * @brief Uložení požadovaného čísla do QByteArray z požadovaných bitů.
 *
 * Implementováno z toho důvodu, že defaultní chování QByteArray je ukládání každé číslice do bajtu. Takže například 1000 se uloží do 4 bajtů, ikdyž by se to vlezlo do 2 bajtů.
 *
 * @param byteCount - Na kolik bitů mají být bity uloženy.
 * @param bits - Číslo konvertováno na bity k uložení.
 * @return Výsledné QByteArray.
 */
QByteArray UdpClient::convertIntToByteArray(int byteCount, QString bits)
{
    QByteArray res;
    for(int b = 0; b < byteCount; b++) {
        int number = 0;
        for(int i = bits.size() - b * 8 - 1, pos = 0; i >= bits.size() - (b + 1) * 8; i--, pos++) {
            if(i >= 0) {
                if(bits.at(i) == '1') {
                    number += pow(2, pos);
                }
            }
        }
        res.append(number);
    }
    return res;
}

/**
 * @brief Odešle se požadavek na začátek skenování a odesílání dat ze serveru do aplikace.
 *
 * Server se chová divně při nastavených hodnotách vzorků, proto je potřeba zkoušet jaký počet je nejlepší (s ohledem na počet nulových měření). Například při požadavku 360 hodnot je pouze cca 120 nenulových, takže je zjíštěno, že je lepší mít například 720 hodnot, protože příchozí data obsahují alespoň 220 hodnot.
 *
 * @param sample - Počet vzorků na měření.
 * @param rate - Jak často se mají hodnoty posílat.
 */
void UdpClient::startGrabbing(int sample, int rate)
{
    QByteArray sampleA = this->convertIntToByteArray(2, QString::number(sample, 2));
    QByteArray rateA = this->convertIntToByteArray(2, QString::number(rate, 2));
    QByteArray msg;
    msg.append(1);
    msg.append(dev_rplidar::Command::StartScan);

    this->stat->sample = sample;
    this->stat->rate = rate;

    this->udpSocket->writeDatagram(msg, this->stat->address, this->stat->serverPort);

    msg.clear();
    msg.append(1);
    msg.append(dev_rplidar::Command::Grab);
    msg.append(sampleA.at(1));
    msg.append(sampleA.at(0));
    msg.append(rateA.at(1));
    msg.append(rateA.at(0));

    this->udpSocket->writeDatagram(msg, this->stat->address, this->stat->serverPort);

    emit this->statisticChange();
}

/**
 * @brief Požadavek na zastavení posílání dat.
 */
void UdpClient::stopGrabbing()
{
    QByteArray msg;
    msg.append(1);
    msg.append(dev_rplidar::Command::Stop);

    this->udpSocket->writeDatagram(msg, this->stat->address, this->stat->serverPort);

    emit this->statisticChange();
}

/**
 * @brief Požadavek na odeslání informací o stavu.
 */
void UdpClient::stateMessage()
{
    QByteArray msg;
    msg.append(1);
    msg.append(dev_rplidar::Command::StateInfo);

    this->udpSocket->writeDatagram(msg, this->stat->address, this->stat->serverPort);

    emit this->statisticChange();
}

/**
 * @brief Funkce pro načtení pouze některých bajtů z datagramu.
 *
 * Takže při požadavku na bajty od 0 do 4, vrátí 0., 1., 2. a 3. bajt.
 *
 * @param start - Index startovního bajtu.
 * @param end - Index koncového bajtu, který už není součástí.
 * @param datagram - Bajtové pole, z kterého se bajty vrací.
 * @return Výsledné QByteArray.
 */
QByteArray UdpClient::getBytesFromDatagram(int start, int end, QByteArray &datagram)
{
    QByteArray a;
    for(int i = start; i < end; i++) {
        a.append(datagram.at(i));
    }
    return a;
}

/**
 * @brief Hodnota v QByteArray konvertována do int.
 * @param array - Pole bajtů.
 * @return Výsledné číslo.
 */
int UdpClient::byteArrayToInt(QByteArray array)
{
    return QString(array.toHex()).toInt(0, 16);
}

/**
 * @brief Čtení příchozích datagramů.
 */
void UdpClient::readPendingDatagrams()
{
    while (udpSocket->hasPendingDatagrams()) {
        QByteArray datagram;
        QByteArray result;
        datagram.resize(udpSocket->pendingDatagramSize());

        udpSocket->readDatagram(datagram.data(), datagram.size());

        if(this->state == ClientState::NotConnected) {
            this->parseModuleList(datagram);
        } else if(this->state == ClientState::ConnectedToServer || this->state == ClientState::ConnectedToLidar || this->state == ClientState::ReceivingData) {
            if(datagram.at(1) == dev_rplidar::Command::StateInfo) {
                this->parseStateMsg(datagram);
            } else if(datagram.at(1) == dev_rplidar::Command::Grab) {
                result = this->getBytesFromDatagram(2, 4, datagram);
                int count = this->byteArrayToInt(result);
                data_t *data = new data_t();
                data->measureCount = count;
                data->measure = new measure_t[count];
                int countReal = 0;
                for(int i = 0; i < count; i++) {
                    // sync_quality, angle_q6_checkbit, distance_q2 = struct.unpack('!BHH', msg[4+i*5:4+i*5+5])
                    result = this->getBytesFromDatagram(4+i*5, 4+i*5+5, datagram);
                    int shift = this->byteArrayToInt(this->getBytesFromDatagram(1, 3, result));
                    shift >>= 1;
                    data->measure[i].angle = shift / 64.0f;
                    data->measure[i].quality = (this->byteArrayToInt(this->getBytesFromDatagram(0, 1, result)) >> 2);
                    data->measure[i].distance = this->byteArrayToInt(this->getBytesFromDatagram(3, 5, result)) / 4.0f;
                    //qDebug() << "measure: " << i << data->measure[i].angle << data->measure[i].quality << data->measure[i].distance;
                    countReal += data->measure[i].distance != 0 ? 1 : 0;
                }

                qDebug() << "Real point count: " << countReal;

                if(this->stat->preprocessData) {
                    this->preProcessor->addNewArea(*data);
                }
            }
        } else {
            qDebug() << datagram;
        }
    }
}


/**
 * @brief Parsování příchozího seznamu modulů dostupných na serveru.
 * @param datagram - Příchozí zpráva.
 */
void UdpClient::parseModuleList(QByteArray &datagram)
{
    if(datagram.at(0) == '\xff') {
        datagram.remove(0, 2);
        QStringList list = QString(datagram).split("##");
        this->modulesCount = list.length();
        this->moduleList = new modules_t[list.length()];
        for(int i = 0; i < list.length(); i++) {
            QStringList item = list[i].split("::");
            this->moduleList[i].position = item.at(0).toInt();
            this->moduleList[i].name = item.at(1);
        }
        this->setState(ClientState::ConnectedToServer);
    } else {
        this->setState(ClientState::NotConnected);
    }
}

/**
 * @brief Parsování informací o stavu.
 * @param datagram - Příchozí zpráva.
 */
void UdpClient::parseStateMsg(QByteArray &datagram)
{
    if(datagram.at(0) == '\x01' && datagram.at(1) == dev_rplidar::Command::StateInfo) {
        datagram.remove(0, 2);
        if(datagram.at(0) == dev_rplidar::State::Waiting) {
            this->setState(ClientState::ConnectedToServer);
        } else if(datagram.at(0) == dev_rplidar::State::Connected) {
            this->setState(ClientState::ConnectedToLidar);
        } else if(datagram.at(0) == dev_rplidar::State::Scanning) {
            this->setState(ClientState::ReceivingData);
        } else if(datagram.at(0) == dev_rplidar::State::Grabbing) {
            this->setState(ClientState::ReceivingData);
        } else if(datagram.at(0) == dev_rplidar::State::Error) {
            qDebug() << "Lidar ERROR!";
        };
    }
}

/**
 * @brief Nastavení stavu UDP.
 * @param state = Stav.
 */
void UdpClient::setState(int state)
{
    this->state = state;

    emit stateChange();
    emit this->statisticChange();
}

/**
 * @brief Vrací uložený seznam modulů.
 * @return Seznam modulů.
 */
QStringList UdpClient::getListOfModules() const
{
    QStringList list;
    for(int i = 0; i < this->modulesCount; i++) {
        list.append(QString::number(moduleList[i].position) + " - " + moduleList[i].name);
    }
    return list;
}

/**
 * @brief Pošle heartbeat na server.
 */
void UdpClient::sendHeartbeat()
{
    if(this->state != ClientState::NotConnected) {
        udpSocket->writeDatagram(this->heartbeatMessage, this->stat->address, this->stat->serverPort);
    }
}

/**
 * @brief Začít příchozí data ukládat a zpracovávat.
 */
void UdpClient::startPreprocessingData()
{
    this->stat->preprocessData = true;
    emit this->statisticChange();
}

/**
 * @brief Nezpracovávat příchozí data.
 */
void UdpClient::stopPreprocessingData()
{
    this->stat->preprocessData = false;
    emit this->statisticChange();
}

/**
 * @brief Vrátí název aktuálního stavu.
 * @return Název aktuálního stavu.
 */
QString UdpClient::getStateName()
{
    if(this->state == ClientState::NotConnected) {
        return "Not Connected";
    } else if(this->state == ClientState::ConnectedToServer) {
        return "Connected to Server";
    } else if(this->state == ClientState::ConnectedToLidar) {
        return "Connected to Lidar";
    } else if(this->state == ClientState::ReceivingData) {
        return "Receiving Data";
    } else {
        return "Not specified";
    }
}
